#!/usr/bin/env python
# coding=utf-8

"""
Definitive module contains methods for geomagnetic definitive data analysis. It is currently part of MARTAS2.0
as it uses the same configuration data as basevalue.
Please note that this module was specifically developed for the Conrad Observatory and some of the methods
will only be useful for very similar procedures.

Definitive analysis methods

| class   |      method     |  version |  tested  |              comment             | manual | *used by |
| ------- |  -------------  |  ------- |  ------- |  ------------------------------- | ------ | ---------- |
|         |  activity_analysis | 2.0.0 |          | analyse K values                 | -      | yearly_analysis_2 |
|         |  blv            |  2.0.0   |          | create blv files                 | -      | yearly_analysis_2 |
|         |  check_update_loc | 2.0.0  |          | replace location with primary pier | -    | yearly_analysis_2 |
|         |  create_rotation_sql |  2.0.0 | yes*  | create SQL string for DB update  | -      | yearly_analysis_1 |
|         |  definitive_min |  2.0.0   |          | create minute dissimination files | -     | yearly_analysis_2 |
|         |  definitive_sec |  2.0.0   |          | create second dissimination files | -     | yearly_analysis_2 |
|         |  pier_diff      |  2.0.0   |   yes    | pier diff relative to mobilesc   | -      | yearly_analysis_1 |
|         |  scalarcomb     |  2.0.0   |   yes    | merge scalar data (min and sec)  | -      | yearly_analysis_1 |
|         |  scalarcorr     |  2.0.0   |   yes    | clean scalar                     | -      | yearly_analysis_1 |
|         |  variocomb      |  2.0.0   |   yes    | merge minute vario datasets      | -      | yearly_analysis_1 |
|         |  variocorr      |  2.0.0   |   yes    | clean, rotate and adopt baseline | -      | yearly_analysis_1 |


"""
import sys
sys.path.insert(1,'/home/leon/Software/MARTAS/') # should be magpy2
sys.path.insert(1,'/home/leon/Software/magpy/') # should be magpy2

import unittest
from copy import deepcopy
from magpy.stream import *
from magpy.core import methods

from martas.core import methods as mm
from martas.app import basevalue
from magpy.core import plot as mp
from magpy.core import activity

from calendar import monthrange


def activity_analysis(config=None, results=None):
    """
    determine quiet,disturbed days
       - get all days with A below 4 and above 30
       - annual diffs - get average A for winter and summer month
    required information:
    """
    print(" ------------------------------------------------- ")
    print(" -- Step 2: Extended information e.g. activity  -- ")
    print(" ------------------------------------------------- ")
    # calc quiet days, sum of quiet days activity etc
    if not config:
        config = {}
    if not results:
        results = {}
    vainstlist = config.get('vainstlist')
    scinstlist = config.get('scinstlist')
    outpath = config.get('outpath',"/tmp")
    plotdir = config.get('plotdir',"/tmp")

    # TODO Extract flagging info on storms and pulsations??
    year = int(config.get('year'))

    actdict = results.get('activity', {})

    outputformats = config.get('outputformats')
    plot = config.get('plot', 'False')

    kupperthreshold = config.get('kupperthreshold')
    klowerthreshold = config.get('klowerthreshold')

    kvals = read(os.path.join(outpath, 'magpy', 'Definitive_kvals_' + str(year) + '.cdf'))

    if plot == 'True':
        mp.plot(kvals, noshow=True)

    orgkvals = kvals.copy()
    dailymeans = kvals.filter(filter_type='flat', filter_width=timedelta(days=1), resampleoffset=timedelta(minutes=780))
    if plot == 'True':
        mp.plot(dailymeans)
    # act < 1:
    quietdays = dailymeans.extract('var1', klowerthreshold, '<')
    print("Amount of quiet days (daily average activity below {}): {}".format(klowerthreshold, quietdays.length()[0]))
    quietdates = [elem.strftime("%Y-%m-%d") for elem in quietdays.ndarray[0]]
    actdict['quiet_days'] = quietdates
    # act > 3.5:
    # disturbeddays = dailymeans.extract('var1',3.5,'>')
    disturbeddays = dailymeans.extract('var1', kupperthreshold, '>')
    print("Amount of disturbed days (daily average activity above {}): {}".format(kupperthreshold,
                                                                                  disturbeddays.length()[0]))
    disturbeddates = [elem.strftime("%Y-%m-%d") for elem in disturbeddays.ndarray[0]]
    actdict['disturbed_days'] = disturbeddates

    ### ######################################################
    print(" --- Mean quiet day field variance --- ")
    ### ######################################################
    # stackStreams

    ### ######################################################
    print(" --- Activity statistics for year and terms --- ")
    ### ######################################################
    # mp.hist([dailymeans],[['var1']])
    pos = KEYLIST.index('var1')
    normit = True
    maxprob = 0.6
    maxN = 1500
    ### #################
    ###   Full year
    plt.subplot(211)
    orgkvals = orgkvals._drop_nans(KEYLIST[pos])
    print(orgkvals.ndarray[pos])
    maxk = int(max(orgkvals.ndarray[pos]))
    print(maxk)
    n, bins, patches = plt.hist(orgkvals.ndarray[pos], maxk, density=normit, facecolor='green', alpha=0.3)
    plt.title(r'$\mathrm{Histogram\ of\ K\ values}\ $')
    if not normit:
        plt.ylabel('N')
        plt.axis([0, 9, 0, maxN])
    else:
        plt.ylabel('Probability')
        plt.axis([0, 9, 0, maxprob])
        plt.text(6, 0.8 * maxprob, r'Year %s' % str(year))
    plt.grid(True)
    ### #################
    ###   Summer
    print(" - Summer")
    cpkvals = orgkvals.copy()
    summervals = orgkvals.trim(starttime='{}-03-21'.format(year), endtime='{}-10-21'.format(year))
    plt.subplot(212)
    """
    n, bins, patches = plt.hist(summervals.ndarray[pos], maxk, normed=normit, facecolor='red', alpha=0.3)
    if not normit:
        plt.ylabel('N')
        plt.axis([0, 9, 0, maxN])
    else:
        plt.ylabel('Probability')
        plt.axis([0, 9, 0, maxprob])
        plt.text(6, 0.8*maxprob, r'Summer term %s' % str(year))
    plt.grid(True)
    """

    ### #################
    ###   Winter
    print(" - Winter")
    cp2kvals = cpkvals.copy()
    # cp2kvals = cp2kvals._drop_nans(KEYLIST[pos])
    wintervals1 = cpkvals.trim(starttime='{}-01-01'.format(year), endtime='{}-03-21'.format(year))
    wintervals2 = cp2kvals.trim(starttime='{}-10-21'.format(year), endtime='{}-12-31'.format(year))
    wintervals = join_streams(wintervals1, wintervals2)
    # plt.subplot(313)
    wintervals = wintervals._drop_nans(KEYLIST[pos])
    # print ("here", len(wintervals.ndarray[pos]))
    print(np.mean(wintervals.ndarray[pos]), np.max(wintervals.ndarray[pos]), np.min(wintervals.ndarray[pos]))
    n, bins, patches = plt.hist((summervals.ndarray[pos], wintervals.ndarray[pos]), maxk, density=normit,
                                color=('red', 'blue'), alpha=0.3)  # int(max(wintervals.ndarray[pos]))
    print("here again")
    plt.xlabel('K')
    if not normit:
        plt.ylabel('N')
        plt.axis([0, 9, 0, maxN])
    else:
        plt.ylabel('Probability')
        plt.axis([0, 9, 0, maxprob])
        plt.text(6, 0.6 * maxprob, r'Summer term %s' % str(year), color='red')
        plt.text(6, 0.8 * maxprob, r'Winter term %s' % str(year), color='blue')
    plt.grid(True)
    if not os.path.exists(plotdir):
        os.makedirs(plotdir)
    plt.savefig(os.path.join(plotdir, 'localk.png'))
    plot = True
    if plot == 'True':
        plt.show()

    ### ######################################################
    print(" --- Average hourly activity statistic distribution --- ")
    ### ######################################################
    """
    streamlist = []
    for day in quietdates:
        select = kvals.copy()
        sd = datetime.strptime(day,"%Y-%m-%d")
        ed = datetime.strptime(day,"%Y-%m-%d")+timedelta(days=1)
        st = select.trim(starttime=sd,endtime=ed)
        streamlist.append(st)
    stacked = stackStreams(streamlist, get='mean')
    print stacked.ndarray
    mp.plot(stacked)
    #stackStreams
    #mp.barchart([dailymeans],[['var1']])
    """
    results['activity'] = actdict
    return results


def blv(config=None, results=None, debug=False):
    print(" -------------------------------------------- ")
    print(" ------- Step 4: Creating BLV files --------- ")
    print(" -------------------------------------------- ")
    if not config:
        config = {}
    if not results:
        results = {}
    outpath = config.get('outpath',"/tmp")
    db = config.get('primaryDB')

    vainstlist = config.get('vainstlist')
    scinstlist = config.get('scinstlist')
    # TODO Extract flagging info on storms and pulsations??
    year = int(config.get('year'))
    blvabb = config.get('blvabb')
    prv = vainstlist[0][:-5]  # primary variometer
    prf = scinstlist[0][:-5]  # primary scalar
    pier = 'A2'
    runmode = 'thirdrun'

    outputformats = config.get('outputformats')
    plot = config.get('plot', 'False')

    print("Loading data")
    basename = '{}_{}_{}_{}'.format(blvabb, prv, prf, pier)  # too save
    flagname = '{}_{}_{}_{}'.format('BLV', prv, prf, pier)  # as flags are stored

    # runmode = 'thirdrun'
    final = read(os.path.join(outpath, 'magpy', 'Definitive_min_mag_{}_{}.cdf'.format(year, runmode)))
    # runmode = 'secondrun'
    absresult = read(os.path.join(outpath, 'magpy', '{}_{}_{}.txt'.format(basename, year, runmode)))
    # Flags have been removed already in yearly_analysis_1

    print(" -- Dropping flags from DB")
    flaglist = absresult.header.get("DataFlags")
    if flaglist:
        print("   -> dropping {} flags in baseline file".format(len(flaglist)))
        absresult = flaglist.apply_flags(absresult, mode='drop')
    if debug:
        print("  Basevalue SensorID:", absresult.header.get('SensorID'))
        print("   -- Dropping basedata flags from DB")
    blvflaglist = db.flags_from_db(absresult.header.get('SensorID'))
    if debug:
        print("   -> {} flags".format(len(blvflaglist)))
    if blvflaglist:
        absresult = blvflaglist.apply_flags(absresult, mode='drop')
    if debug:
        print("   -- Dropping basedata flags from file: {}".format(config.get('diflagfile')))
    fiflaglist = flagging.load(config.get('diflagfile', ''), sensorid=absresult.header.get('SensorID'))
    if debug:
        print("   -> {} flags".format(len(fiflaglist)))
    if fiflaglist:
        absresult = fiflaglist.apply_flags(absresult, mode='drop')
    print(" -- Remaining abolsutes: {}".format(absresult.length()[0]))

    # Add pier diff to delta F values

    print("Calculating deltaF")
    final = final.delta_f()
    if plot == 'True':
        mp.plot(final)
    pos = KEYLIST.index('df')
    # print "Mean delta F:", final.ndarray[pos]

    print("Determine means")
    pos = KEYLIST.index('df')
    final = final.xyz2hdz()
    meanh = final.mean('x')
    meanf = final.mean('f')

    print("Move delta F column to correct position")
    df = final._get_column('df')
    final = final._put_column(df, 'f')
    print("Filtering daily means")
    dailymean = final.dailymeans(keys=['x', 'y', 'z', 'f', 'df'])
    df = dailymean._get_column('f')
    dailymean = dailymean._put_column(df, 'df')

    if plot == 'True':
        mp.plot(dailymean, variables=['f', 'df'])
    # print "Dailymean", dailymean.length(), dailymean.header

    # Modify the absresult:
    # a) add pierdiff to baseline
    val = db.select('DataDeltaValues', 'DATAINFO', 'DataID="{}"'.format(scinstlist[0]))[0]
    dic = string2dict(val, typ='listofdict')
    # get all inputs with starttime <= end of year and endtime either missing or >= start of year
    dflist = []
    for el in dic:
        stt = date2num(datetime(int(year), 1, 1))
        ett = date2num(datetime(int(year), 12, 31))
        st = float(el.get('st'))
        et = float(el.get('et', ett))
        if st <= ett and et > stt:
            dflist.append(el.get('f', np.nan))
    dflist = list(set(dflist))
    if len(dflist) > 1 or len(dflist) < 0:
        print("IMPORTANT: check delta F values for this year")
    else:
        deltaF = float(dflist[0])
        print("Extracted a deltaF of {} from the database for {}".format(deltaF, scinstlist[0]))

    print(" -- Remaining abolsutes: {}".format(dailymean.length()[0]))

    print(" -- Absinfo looks like: {}".format(final.header.get('DataAbsInfo')))
    absresult.header['StationID'] = 'WIC'
    absresult.write(os.path.join(outpath, 'IAF'), coverage='all', format_type='BLV', deltaF=deltaF, diff=dailymean,
                    year=str(year), meanh=meanh, meanf=meanf,
                    absinfo=final.header.get('DataAbsInfo'))  # add daily means of deltaf

    return results


def check_update_location(header, db, reference='A2'):
    # Get A2 Position
    lon = db.select('PierLong', 'PIERS', 'PierID = "{}"'.format(reference))[0]
    lat = db.select('PierLat', 'PIERS', 'PierID = "{}"'.format(reference))[0]
    try:
        if isinstance(lon, str):
            lon = lon.replace(",", ".")
        if isinstance(lat, str):
            lat = lat.replace(",", ".")
        lon = float(lon)
        lat = float(lat)
    except:
        print("Problem with database content of PIERS")
        lon = None
        lat = None
    if lat and not (header.get('DataAcquisitionLatitude') == lat):
        print(
            " Found {} Location which differs from file content - updating DataAcquisitionLatitude: {} -> {}".format(
                reference, header['DataAcquisitionLatitude'], lat))
        header['DataAcquisitionLatitude'] = np.round(lat, 5)
    if lon and not (header.get('DataAcquisitionLongitude') == lon):
        print(
            " Found {} Location which differs from file content - updating DataAcquisitionLomgitude: {} -> {}".format(
                reference, header['DataAcquisitionLongitude'], lon))
        header['DataAcquisitionLongitude'] = np.round(lon, 5)

    return header


def create_rotation_sql(rotangledict, config=None, debug=False):

    if not config:
        config = {}
    vainstlist = config.get('vainstlist')
    year = config.get('year')
    db = config.get('primaryDB')
    rotation = False
    meanrotangle = 0.0
    sqlstatement = []
    print ("IMPORTANT: update database with rotation")
    print ("----------------------------------------")
    for vainst in vainstlist:
        # get existing rotangle data
        header = db.fields_to_dict(vainst)
        exist = header.get('DataRotationAlpha','')
        existbeta = header.get('DataRotationBeta','')
        # obtain new
        rotanglelist = rotangledict.get(vainst,[])
        meanrotangle = np.median(rotanglelist,axis=0)

        if not exist == '':
            val = '{},{}_{:.3f}'.format(exist,year,meanrotangle[0])
        else:
            val = '{}_{:.3f}'.format(year,meanrotangle[0])
        updatestr = "UPDATE DATAINFO SET DataRotationAlpha='{}' WHERE SensorID LIKE '{}%';".format(val,vainst[:-5])
        sqlstatement.append(updatestr)
        print (updatestr)
        if not existbeta == '':
            val = '{},{}_{:.3f}'.format(existbeta,year,meanrotangle[1])
        else:
            val = '{}_{:.3f}'.format(year,meanrotangle[1])
        updatestr = "UPDATE DATAINFO SET DataRotationBeta='{}' WHERE SensorID LIKE '{}%';".format(val,vainst[:-5])
        sqlstatement.append(updatestr)
        print (updatestr)
    print ("----------------------------------------")
    return sqlstatement


def definitivefiles_min(config=None, results=None, runmode="thirdrun", headdict=None):
    if not config:
        config = {}
    if not results:
        results = {}
    if not headdict:
        headdict = {'DataSensorOrientation' : 'hdz', 'StationInstitution' : ' GSA - GeoSphere Austria'}

    vainstlist = config.get('vainstlist')
    scinstlist = config.get('scinstlist')
    outpath = config.get('outpath',"/tmp")
    db = config.get('primaryDB')

    year = int(config.get('year'))

    outputformats = config.get('outputformats')
    # pierlist = config.get('pierlist')
    plot = config.get('plot', 'False')
    # blvabb = config.get('blvabb')
    monthres = {}

    print(" ---------------------------------------------------- ")
    print(" ---- Definitive minute and calculate k values  ----- ")
    print(" ---------------------------------------------------- ")
    print(" ------------------------------ ")
    print(" a) creating minute magpy archive ")
    # expc = 365*1440
    expc = 0
    for i in range(0, 12):
        expc += monthrange(year, int(i + 1))[1] * 24 * 60
    merge = read(os.path.join(outpath, 'magpy', 'Definitive_min_mag_{}_{}.cdf'.format(year, runmode)))
    pos = KEYLIST.index('x')
    col = np.asarray(merge.ndarray[pos].astype(float))
    count = col.size - np.count_nonzero(np.isnan(col))
    print("Expected: {b}, Contained: {c}, Coverage: {d}".format(b=expc, c=count, d=float(count) / float(expc) * 100.))
    monthres['vario-expected'] = expc
    monthres['vario-count'] = count
    monthres['vario-coverage'] = float(count) / float(expc) * 100.
    pos = KEYLIST.index('f')
    col = np.asarray(merge.ndarray[pos].astype(float))
    count = col.size - np.count_nonzero(np.isnan(col))
    print("Expected: {b}, Contained: {c}, Coverage: {d}".format(b=expc, c=count, d=float(count) / float(expc) * 100.))
    monthres['scalar-expected'] = expc
    monthres['scalar-count'] = count
    monthres['scalar-coverage'] = float(count) / float(expc) * 100.

    results['coverage-min'] = monthres
    # Adding special header information (not contained in data base):
    # For IAF
    merge.header['DataConversion'] = merge.mean('x') / 3438. * 10000.
    merge.header['DataQuality'] = 'IMAG'
    # For IMAGCDF
    merge.header['DataPublicationLevel'] = 4  # 4 corresponds to definitive
    # TODO remove - they are just here because I forgot them in the database
    for el in headdict:
        merge.header[el] = headdict.get(el)

    merge = merge._convertstream('hdz2xyz')
    merge.header['col-f'] = 'F'
    print(" Checking Location information and eventually update this information to A2")
    merge.header = check_update_location(merge.header, db, reference='A2')

    merge.write(os.path.join(outpath, 'magpy'), filenamebegins='Definitive_min_mag_' + str(year), dateformat='%Y',
                coverage='all', format_type='PYCDF')

    print ("Now producing output formats: {}. Header information looks like: {}".format(outputformats, merge.header))

    if plot == 'True':
        mp.plot(merge)

    if 'MagPyK' in outputformats:
        print(" ------------------------------ ")
        print(" b) running k values determination ")
        #print ("Header", merge.header)
        # TODO - this is not perfectly correct. We need at least one day from the past and the following year
        print ("ATTENTION : IF THE TIMERANGE OF MERGE IS TO SHORT THEN KVALS ARE MISSING", merge.timerange())
        kvals = activity.K_fmi(merge, K9_limit=500, longitude=15.87)
        #kvals = merge.k_fmi()
        kvals.write(os.path.join(outpath, 'magpy'), filenamebegins='Definitive_kvals_' + str(year), dateformat='%Y',
                    coverage='all', format_type='PYCDF', skipcompression=True)

    print(" ------------------------------ ")
    print(" c) Writing dissemination files ")
    # Get deltaF values
    final = merge.delta_f()
    final = final.get_gaps()

    backup = deepcopy(final.header)
    # temporary change 2020 as proj seems to be wrongly converting longitudes on laptop
    ##final.header['DataAcquisitionLongitude']  = -35130

    if 'IAGA' in outputformats:
        print("Rounding location accuracy")
        final.header['DataAcquisitionLatitude'] = np.round(float(final.header.get('DataAcquisitionLatitude')), 5)
        final.header['DataAcquisitionLongitude'] = np.round(float(final.header.get('DataAcquisitionLongitude')), 5)
        print("Writing IAGA files ...")
        print ("Lat", final.header.get('DataAcquisitionLatitude') )
        print("Header info on IAGA code:", final.header.get('StationIAGAcode'), final.header.get('StationID'))
        # IAGA: (only minute data)cd
        final.write(os.path.join(outpath, 'IAGA', 'minute'), filenamebegins=final.header.get('StationIAGAcode'),
                    filenameends='dmin.min', dateformat='%Y%m%d', format_type='IAGA')

    if 'IMAGCDF' in outputformats:
        print("Writing IMAGCDF minute files ...")
        # ImagCDF:
        final.header = deepcopy(backup)
        final.write(os.path.join(outpath, 'ImagCDF'), coverage='year', format_type='IMAGCDF')

    if 'IAF' in outputformats and 'MagPyK' in outputformats:
        print("Writing IAF minute files ...")
        # IAF:  # Changing institution name as only first 4 digits are used in IAF
        ##### Remove DKA file if you rewrite IAF
        final.header = deepcopy(backup)
        final.write(os.path.join(outpath, 'IAF'), coverage='month', format_type='IAF', kvals=kvals, debug=True)
        final.write(os.path.join(outpath, 'IAF'), kind='A', format_type='IYFV')

    return results


def definitivefiles_sec(config=None, startmonth=0, endmonth=12, results=None, runmode="thirdrun", headdict=None):
    # Create monthly one second imagcdf files
    # Add characteristics to a results dictionary
    print(" ------------------------------------------------ ")
    print(" ---------- Create Monthly ImagCDF's  ----------- ")
    print(" ------------------------------------------------ ")
    if not config:
        config = {}
    if not results:
        results = {}
    if not headdict:
        headdict = {'DataReferences' : 'cobs.geosphere.at', 'StationInstitution' : 'GeoSphere Austria', 'DataStandardLevel' : 'Full'}

    vainstlist = config.get('vainstlist')
    scinstlist = config.get('scinstlist')
    outpath = config.get('outpath',"/tmp")
    db = config.get('primaryDB')
    year = int(config.get('year'))
    addflags = config.get("addflags", False)

    outputformats = config.get('outputformats')
    # pierlist = config.get('pierlist')
    plot = config.get('plot', 'False')
    # blvabb = config.get('blvabb')
    monthsum = {}

    prv = vainstlist[0]  # primary variometer
    prf = scinstlist[0]  # primary scalar

    for i in range(startmonth, endmonth):
        monthres = {}
        month = str(i + 1).zfill(2)
        nextmonth = str(i + 2).zfill(2)
        nextyear = year
        if nextmonth == '13':
            nextyear = year + 1
            nextmonth = '01'
        expc = monthrange(year, int(i + 1))[1] * 24 * 3600
        print("Dealing with month: {}".format(month))
        vaname = "{}_vario_sec_{}_{}{}.cdf".format(prv, runmode, year, month)
        print("Loading variation data: {}".format(vaname))
        print("----------------------")
        vario = read(os.path.join(outpath, 'magpy', vaname))
        print("  -- loaded data")
        if plot == 'True':
            mp.tsplot(vario)
        flags = vario.header.get("DataFlags")
        if flags:
            vario = flags.apply_flags(vario, mode="drop")
        print("  -- removed flagged")
        vario = vario.bc()
        print("  -- corrected baseline")
        vario = vario._convertstream('hdz2xyz')
        print("  -- converted to xyz")
        # TODO - join if several varios
        # main = vario.copy()
        # else:
        # main = join_streams(main, vario)
        print("  -- dropping columns except field and T")
        vario = vario._drop_column('var1')
        vario = vario._drop_column('var2')
        vario = vario._drop_column('var5')
        vario = vario._drop_column('str1')
        pos = KEYLIST.index('x')
        col = np.asarray(vario.ndarray[pos])
        count = col.size - np.count_nonzero(np.isnan(col))
        print("  Month {a} ---  Expected: {b}, Contained: {c}, Coverage: {d} %".format(a=month, b=expc, c=count,
                                                                                       d=float(count) / float(
                                                                                           expc) * 100.))
        monthres['vario-expected'] = expc
        monthres['vario-count'] = count
        monthres['vario-coverage'] = float(count) / float(expc) * 100.
        scname = "{}_scalar_oc_sec_{}_{}{}.cdf".format(prf, runmode, year, month)
        print("Loading scalar data:", scname)
        print("----------------------")
        scalar = read(os.path.join(outpath, 'magpy', scname))
        sr = scalar.samplingrate()
        flags = scalar.header.get("DataFlags")
        if flags:
            scalar = flags.apply_flags(scalar, mode="drop")
        print("  -- removed flagged")
        print(scalar.length())
        print("  -- droping duplicates")
        scalar = scalar.removeduplicates()
        print(scalar.length())
        print("  -- rersampling at 1 sec")
        scalar = scalar.resample(['f'], period=1)
        expc = monthrange(year, int(i + 1))[1] * 24 * int(3600 / float(sr))
        # flaglist = db2flaglist(db,scinst[:-5])
        # scalar = scalar.flag(flaglist)
        # print (scalar.length())
        # scalar = scalar.remove_flagged()
        pos = KEYLIST.index('f')
        col = np.asarray(scalar.ndarray[pos].astype(float))
        count = col.size - np.count_nonzero(np.isnan(col))
        ### Get amount of expected and missing data
        print("  Month {a} ---  Expected: {b}, Contained: {c}, Coverage: {d} %".format(a=month, b=expc, c=count,
                                                                                       d=float(count) / float(
                                                                                           expc) * 100.))
        monthres['scalar-expected'] = expc
        monthres['scalar-count'] = count
        monthres['scalar-coverage'] = float(count) / float(expc) * 100.
        monthsum[month] = monthres
        if not sr == 1:
            print("Sampling rate of scalar data: {}".format(sr))
        #merge = join_streams(vario, scalar)
        merge = merge_streams(vario, scalar, keys=['f'])
        print("Merge lenght", merge.length())

        merge = merge.get_gaps()
        # For IMAGCDF
        merge.header['DataPublicationLevel'] = 4  # 4 corresponds to definitive
        for el in headdict:
            merge.header[el] = headdict.get(el)

        # Site Location
        print(" Checking Location information and eventually update this information to A2")
        merge.header = check_update_location(merge.header, db, reference='A2')

        if 'IMAGCDF' in outputformats:
            # savedata to IMAGCDF onesecond standard
            print("Writing ImagCDF second")
            merge.write(os.path.join(outpath, 'ImagCDF'), coverage='month', format_type='IMAGCDF', addflags=addflags)

        if 'IAGA' in outputformats:
            # savedata to IAGA02 onesecond standard
            print("Rounding location accuracy")
            merge.header['DataAcquisitionLatitude'] = np.round(float(merge.header.get('DataAcquisitionLatitude')), 5)
            merge.header['DataAcquisitionLongitude'] = np.round(float(merge.header.get('DataAcquisitionLongitude')), 5)
            print("Writing IAGA second")
            merge.write(os.path.join(outpath, 'IAGA', 'second'), format_type='IAGA')

        print("Coverage dictionary now looks like: {}".format(monthsum))

    results['coverage'] = monthsum
    return results


def pier_diff(runmode, config=None, debug=False):
    """
    REQUIREMENT:
        python3

    DESCRIPTION:
     Check DI differnces for all piers:
     Load definite data for primary instruments (minute resolution is fine)
     Convert to IDF
     Then load DI data for all piers and calculate average diffs relative to definite data
     -> Checks diffs (should be zero for A2)
    """
    if not config:
        config = {}

    sqlstatement = []

    print ("------------------------------------")
    print ("Pier differences")
    print ("------------------------------------")

    vainstlist = config.get('vainstlist')
    scinstlist = config.get('scinstlist')
    pierlist = config.get('pierlist')
    year = int(config.get('year'))
    plot = config.get('plot','False')
    blvabb = config.get('blvabb')
    mobileinst = config.get('mobileinst')
    outpath = config.get('outpath',"/tmp")
    db = config.get('primaryDB')


    prv = vainstlist[0][:-5]  # primary variometer
    prf = scinstlist[0][:-5]  # primary scalar

    # Load definite data dependent on runmode
    if runmode in ['secondrun']:
        # Read one minute definite data
        definitive = read(os.path.join(outpath,'magpy','Definitive_min_mag_{}_{}.cdf'.format(year,runmode)))
        print ("Definitive one minute data with {} data points loaded".format(definitive.length()[0]))
    else:
        print ("Runmode not valid aborting")
        return []
    defi = definitive.hdz2xyz()
    defi = defi.xyz2idf()

    print ("Loading F data from mobile sensor")
    mobilef = read(os.path.join(outpath,'magpy','{}_0001_scalar_oc_min_*'.format(mobileinst)))
    print ("    -> got {} datapoints".format(mobilef.length()[0]))
    print (" - getting standard flags for this sensor")
    #from file
    flagfilename = os.path.join(outpath,'magpy','{}_flags.json'.format(mobileinst))
    mobflaglist = flagging.load(flagfilename, sensorid=mobileinst)
    #from DB
    mobflaglist2 = db.flags_from_db(mobileinst)
    if len(mobflaglist) > 0 and len(mobflaglist2) > 0:
        mobflaglist = mobflaglist.join(mobflaglist2)
    elif len(mobflaglist2) > 0:
        mobflaglist = mobflaglist2
    print (" - found {} flags".format(len(mobflaglist)))
    # Applying all flags
    if mobflaglist:
        mobilef = mobflaglist.apply_flags(mobilef, mode='drop')
    print (" - getting pier flags for this sensor")
    pierfilename = os.path.join(outpath,'magpy','{}_piers.json'.format(mobileinst))
    pierflaglist = flagging.load(pierfilename, sensorid=mobileinst)
    if not len(pierflaglist) > 0:
        print (" --------------------------------------")
        print (" Need to access the pier flaglist to perform this method - aborting")
        print (" --------------------------------------")

    def get(selfi, searchdict, combine='and'):
        """
        DESCRIPTION:
            extract data from flaglist

        EXAMPLE:
            newflaglist = flaglist.get({'comment':'lightning'})
        """

        extractedflaglist = []

        for idx,searchcrit in enumerate(searchdict):
            if combine == 'and' and idx >= 1:
                flaglist = extractedflaglist
            elif combine == 'and' and idx == 0:
                flaglist = selfi
            else: # or
                flaglist = selfi
            #print (searchcrit, searchdict[searchcrit])
            pos = 4 #self.FLAGKEYS.index('comment')
            fl = [el for el in flaglist if searchdict[searchcrit] in el[pos]]
            extractedflaglist.extend(fl)

        return extractedflaglist

    for i in range(1,18):
        pier = 'A{}'.format(i)
        if i == 17:
            pier = 'H1'
        print ("-------------------------------------------")
        print (" Dealing with pier {}".format(pier))
        print ("-------------------------------------------")

        ad, adstd = '',''
        ai, aistd = '',''
        af, afstd = '',''

        print ("############################")
        print (" - Getting existing values from data base")

        # Delta Dictionary in PIERS
        #
        val= db.select('DeltaDictionary','PIERS','PierID like "{}"'.format(pier))[0]
        dic = string2dict(val) #, typ='olddeltadict')
        # Get A2 values
        a2dic = dic['A2']   # append new values here (a2dic[year] = newvaluedict; dic['A2'] = a2dic)
        # get last year for each value
        existdelta = []
        for elem in ['deltaD','deltaI','deltaF']:
            # get years when elem was determined
            years = [int(ye) for ye in a2dic if not a2dic[ye].get(elem,'') == '']
            if len(years) > 0:
                value = a2dic.get(str(max(years))).get(elem,'')
                line = "{} ({}): {}".format(elem,max(years),value)
            else:
                line = "{} ([]): never determined".format(elem)
            existdelta.append(line)
        if debug:
            print("   Found existing deltas:", existdelta)

        print ("############################")
        print (" - Extract F values from mobile sensor for pier {}".format(pier))

        #currentpier = get(pierflaglist,{'comment':pier})
        currentpier = pierflaglist.select(parameter='comment', values=pier, identical=True, debug=True)
        print ("Current pier:", len(currentpier))
        if len(currentpier) > 0:
            print (" - found valid data range ... ")
            arsum = [[]]*24
            for el in currentpier.flagdict:
                fd = currentpier.flagdict
                flel = fd.get(el)
                if debug:
                    print ("    found flag element:", flel)
                ar = mobilef._select_timerange(starttime=flel.get('starttime'), endtime=flel.get('endtime'))
                if debug:
                     print ("    obtained {} data points from mobile sensor".format(len(ar[0])))
                if not len(arsum[0]) > 0:
                    arsum = ar
                    print (arsum)
                else:
                    for idx,el in enumerate(ar):
                        arsum[idx] = np.concatenate((arsum[idx],ar[idx]))
                    #print (len(arsum[0]))
            print (" - got {} data points".format(len(arsum[0])))
            pierts = DataStream([],mobilef.header,arsum)
            pierts = pierts.get_gaps()
            #mp.plot(pierts)
            if debug:
                print ("  Definitive:", definitive.timerange(), definitive.samplingrate())
                print ("  Pierts:", pierts.timerange(), pierts.samplingrate())

            fdiff = subtract_streams(definitive, pierts)
            fdiff = fdiff.get_gaps()
            if debug:
                print ("  diffs between definitive and piers:", len(fdiff))

            fo = flagging.flag_outlier(fdiff)
            if fo:
                fdiff = fo.apply_flags(fdiff, 'drop')
                print ("  dropping outliers:", len(fo))

            if plot == 'True':
                mp.tsplot(fdiff, title=pier)
            fdiff = fdiff._drop_nans('f')

            af,afstd = fdiff.mean('f',meanfunction='median',std=True,percentage=5)
            print (" - F-results for {}: {}+/-{}".format(pier, af,afstd))

        if pier in pierlist:
            print ("############################")
            print (" - Running directional delta analysis for pier {}".format(pier))
            vals1 = defi.copy()

            blvname = '{}_{}_{}_{}_{}_{}.txt'.format(blvabb,prv,prf,pier,year,runmode)
            #delete
            #blvname = '{}-step7.txt'.format(pier)
            print (" - Loading {}".format(blvname))
            try:
                blvdata = read(os.path.join(outpath,'magpy',blvname))
                #delete
                #blvdata = read(os.path.join('/home/leon/Dropbox/Daten/IAGA/DI-files',blvname))
                print (" - Got {} data points in BLV file".format(blvdata.length()[0]))
            except:
                blvdata = DataStream()

            if blvdata.length()[0] > 0:
                print (" - Analyzing {}".format(blvname))
                print (" - Determining dD, dI (and dF) in comparison to Definite_min...cdf file (based on A2):")
                print (" - Trimming original data to overlapping range")
                # remove flags first
                print("  Flagging baseline ...")
                flaglist = blvdata.header.get("DataFlags")
                if flaglist:
                    print("   -> dropping {} flags in baseline file".format(len(flaglist)))
                    blvdata = flaglist.apply_flags(blvdata, mode='drop')
                if debug:
                    print("  Basevalue SensorID:", blvdata.header.get('SensorID'))
                    print("   -- Dropping basedata flags from DB")
                blvflaglist = db.flags_from_db(blvdata.header.get('SensorID'))
                if debug:
                    print("   -> {} flags".format(len(blvflaglist)))
                if blvflaglist:
                    blvdata = blvflaglist.apply_flags(blvdata, mode='drop')
                if debug:
                    print("   -- Dropping basedata flags from file: {}".format(config.get('diflagfile')))
                fiflaglist = flagging.load(config.get('diflagfile', ''), sensorid=blvdata.header.get('SensorID'))
                if debug:
                    print("   -> {} flags".format(len(fiflaglist)))
                if fiflaglist:
                    blvdata = fiflaglist.apply_flags(blvdata, mode='drop')
                vals2 = blvdata.copy()
                #vals2 = vals2.idf2xyz()
                #mp.plot(vals1)

                diff = subtract_streams(vals1,vals2, keys=['x','y','z','f'])
                if diff.length()[0] > 2:
                    print (" - difference values: {}".format(diff.length()[0]))
                    fl = flagging.flag_outlier(diff, keys=['x','y','z','f'], markall=True)
                    if fl:
                        diff = fl.apply_flags(diff, mode='drop')
                    if plot == 'True':
                        mp.tsplot(diff)
                    print (" - remaining after quick outlier removal: {}".format(diff.length()[0]))
                    ai,aistd = diff.mean('x',meanfunction='median',std=True,percentage=50)
                    ad,adstd = diff.mean('y',meanfunction='median',std=True,percentage=50)
                    az,azstd = diff.mean('z',meanfunction='median',std=True,percentage=50)

        print ("############################")
        print ("Delta values results for pier {}".format(pier))
        valdic = {}
        for idx,elem in enumerate(['deltaD','deltaI','deltaF']):
            print ("{}:".format(elem))
            print (" -> Last determination: {}".format(existdelta[idx]))
            short = elem[-1].lower()
            val = eval("a{}".format(short))
            stdval = eval("a{}std".format(short))
            if short in ['i','d'] and not ad == '' and not ai == '' and np.abs(stdval) > np.abs(val):
                # Significance test
                print (" -> New value (not significant): {} +/- {}".format(val,stdval))
                valdic[elem] = '0.00001'
            else:
                print (" -> New value: {} +/- {}".format(val,stdval))
                if not str(val) == '':
                    if short in ['d','i']:
                        strval = "{:.5f}".format(val)
                    else:
                        strval = "{:.2f}".format(val)
                    valdic[elem] = strval

        if valdic:
            a2dic[str(year)] = valdic
            dic['A2']  = a2dic

        command = dict2string(dic)
        if debug:
            print (command)
        sqlline = "UPDATE PIERS SET DeltaDictionary='{}' WHERE PierID='{}';".format(command,pier)
                #sqlstatement.append(sqlline)
        dic = string2dict(command,typ='dictionary')
        if debug:
            print (dic)
        sqlstatement.append(sqlline)

    print (sqlstatement)
    return sqlstatement

def scalarcomb(runmode, config=None, startmonth=-1, endmonth=13, debug=False):
    """
    DESCRIPTION
        create a merged scalar data file
    TODO
       does not yet work properly - a manual analysis using xmagpy works fine with merge
    """
    if not config:
        config = {}
        return

    outpath = config.get('outpath',"/tmp")

    print (" ----------------------------------------- ")
    print (" --- Combining F measurements --- ")
    print (" ----------------------------------------- ")
    ### !! Run from -1 to 13 to cover the time range of abolutes:
    year = config.get('year')
    scinstlist = config.get('scinstlist')

    f, dt = [],[]
    combminute = False
    combsecond = False
    for i in range(startmonth,endmonth):
        # Extract time range (monthly junks)
        startyear = year
        nextyear = year
        month = str(i+1).zfill(2)
        if i == -1:
           startyear = year-1
           month = str(i+13).zfill(2)
        if i+1 >= 13:
            startyear = year+1
            month = str(i-11).zfill(2)
        nextmonth = str(i+2).zfill(2)
        if i+2 >= 13:
            nextyear = year+1
            nextmonth = str(i-10).zfill(2)

        dstarttime=datetime.strptime(str(startyear)+'-'+month+'-01',"%Y-%m-%d")
        dendtime=datetime.strptime(str(nextyear)+'-'+nextmonth+'-01',"%Y-%m-%d")
        print ("Dealing with data between {} and {}".format(dstarttime,dendtime))
        refmin = DataStream()
        ref = DataStream()

        for idx,scinst in enumerate(scinstlist):
            if idx == 0:
                print ("-------------------------------------")
                print ("Base data: {}".format(scinst))
                print ("-------------------------------------")
                if not combminute:
                    refmin = read(os.path.join(os.path.join(outpath,'magpy','{}_scalar_oc_min_{}_{}.cdf'.format(scinst,runmode,year))))
                    print ("Got {} data points in reference min file".format(refmin.length()[0]))
                    print ("Drop line with nan values")
                    refmin = refmin.get_gaps()
                    for el in DataStream().KEYLIST:
                        if not el in ['time','f']:
                            refmin = refmin._drop_column(el)
                    refmin = refmin._drop_nans('f')
                if not combsecond and os.path.isfile(os.path.join(os.path.join(outpath,'magpy','{}_scalar_oc_sec_{}{}{:02d}.cdf'.format(scinst,runmode,dstarttime.year,dstarttime.month)))):
                    ref = read(os.path.join(os.path.join(outpath,'magpy','{}_scalar_oc_sec_{}{}{:02d}.cdf'.format(scinst,runmode,dstarttime.year,dstarttime.month))))
                    print ("Got {} data points in reference sec file".format(ref.length()[0]))
                    ref = ref.get_gaps()
                    for el in DataStream().KEYLIST:
                        if not el in ['time','f']:
                            ref = ref._drop_column(el)
                    # Drop flagged data
                    print ("Drop flagged")
                    print (ref.length(), ref.timerange())
                    flaglist = ref.header.get("DataFlags")
                    if flaglist:
                        ref = flaglist.apply_flags(ref, mode='drop')
                    ref = ref.remove_flagged()
                    print ("Resampling data")
                    print (ref.length())
                    print (ref.timerange())
                    ref = ref.resample(['f'],period=1)
                    print ("Drop line with nan values")
                    ref = ref._drop_nans('f')
            else:
                print ("-------------------------------------")
                print ("Adding data from {}".format(scinst))
                print ("-------------------------------------")
                if not combminute:
                    if os.path.isfile(os.path.join(outpath,'magpy','{}_scalar_oc_min_{}_{}.cdf'.format(scinst,runmode,year))):
                        addmin = read(os.path.join(outpath,'magpy','{}_scalar_oc_min_{}_{}.cdf'.format(scinst,runmode,year)))
                        print ("Got {} data points in min file".format(addmin.length()[0]))
                        addmin = addmin.get_gaps()
                        for el in DataStream().KEYLIST:
                            if not el in ['time', 'f']:
                                addmin = addmin._drop_column(el)
                        print ("Drop line with nan values")
                        addmin = addmin._drop_nans('f')
                        print ("Merging timeseries")
                        refmin = join_streams(refmin,addmin)
                        refmin = merge_streams(refmin,addmin)
                        print ("Reference now contains {} data points (min)".format(refmin.length()[0]))
                if not combsecond:
                    if os.path.isfile(os.path.join(outpath,'magpy','{}_scalar_oc_sec_{}{}{:02d}.cdf'.format(scinst,runmode,dstarttime.year,dstarttime.month))):
                        add = read(os.path.join(outpath,'magpy','{}_scalar_oc_sec_{}{}{:02d}.cdf'.format(scinst,runmode,dstarttime.year,dstarttime.month)))
                        print ("Got {} data points in sec file".format(add.length()[0]))
                        add = add.get_gaps()
                        for el in DataStream().KEYLIST:
                            if not el in ['time', 'f']:
                                add = add._drop_column(el)
                        # Get sampling rate and only continue if sr <= 1
                        sr = add.samplingrate()
                        print ("Sampling rate : {} sec".format(sr))
                        if not sr > 1.01:
                            print ("sampling rate OK - continue")
                            # Drop flagged data
                            print ("Drop flagged")
                            flags = add.header.get("DataFlags")
                            if flags:
                                add = flags.apply_flags(add, mode='drop')
                            print ("Resampling data")
                            add = add.resample(['f'],period=1)
                            print ("Drop line with nan values")
                            add = add._drop_nans('f')
                            #print ("Checking subtract")
                            #subtract = subtractStreams(ref,add)
                            #flaglist = subtract.flag_outlier(threshold=3,returnflaglist=True)
                            print ("Merging timeseries")
                            ref = join_streams(ref,add)
                            ref = merge_streams(ref,add)
                            print ("Reference now contains {} data points (sec)".format(ref.length()[0]))
                        else:
                            print ("sampling rate too large for second combination")
        if refmin.length()[0] > 0:
            print ("Minute data combined")
            combminute = True
            refmin = refmin.get_gaps()
            refmin.write(os.path.join(outpath,'magpy'),filenamebegins='CobsF_min_{}'.format(runmode),coverage='all',mode='overwrite',format_type='PYCDF')
            #save to file
        if ref.length()[0] > 0:
            ref = ref.get_gaps()
            print ("Second data combined for {}".format(dstarttime.month))
            ref.write(os.path.join(outpath,'magpy'),filenamebegins='CobsF_sec_{}_'.format(runmode),dateformat='%Y%m',coverage='month',mode='replace',format_type='PYCDF')


def scalarcorr(runmode, config=None, startmonth=-1, endmonth=13, flagfile=None, skipsec=False, debug=False):
    if debug:
        print (" ----------------------------------------- ")
        print (" --- Checking and Filtering F --- ")
        print (" ----------------------------------------- ")
    ### !! Run from -1 to 13 to cover the time range of abolutes:
    if not config:
        config = {}
        return
    year = config.get('year')
    scinstlist = config.get('scinstlist')
    testdate = config.get('testdate',"")

    sourcepath = os.path.join(config.get('base'), 'archive', config.get('obscode'))
    db = config.get('primaryDB')
    outpath = config.get('outpath',"/tmp")

    if runmode == 'secondrun':
        scinstlist.append("{}_0001".format(config.get('mobileinst')))

    #### If a testdate is provided (single day)
    tdate = None
    if testdate:
        #extract date and month
        tdate = datetime.strptime(testdate,"%Y-%m-%d")
        if not tdate.year == year:
            print ("Year not fitting - ignoring testdate")
            tdate = None
        else:
            startmonth = tdate.month-1
            endmonth = tdate.month

    f, dt = [],[]
    for i in range(startmonth,endmonth):
        # Extract time range (monthly junks)
        startyear = year
        nextyear = year
        month = str(i+1).zfill(2)
        if i == -1:
           startyear = year-1
           month = str(i+13).zfill(2)
        if i+1 >= 13:
            startyear = year+1
            month = str(i-11).zfill(2)
        nextmonth = str(i+2).zfill(2)
        if i+2 >= 13:
            nextyear = year+1
            nextmonth = str(i-10).zfill(2)

        for scinst in scinstlist:
            if debug:
                print ("-------------------------------------")
                print ("Running analysis for {}".format(scinst))
                print ("-------------------------------------")
                print ("Using Scalar data from path: {}".format(os.path.join(sourcepath,scinst[:-5],scinst,'*')))

            if not tdate:
                dstarttime=datetime.strptime(str(startyear)+'-'+month+'-01',"%Y-%m-%d")
                dendtime=datetime.strptime(str(nextyear)+'-'+nextmonth+'-01',"%Y-%m-%d")
            else:
                dstarttime=tdate
                dendtime=tdate+timedelta(days=1)
            if debug:
                print ("  Dealing with data between {} and {}".format(dstarttime,dendtime))

            try:
                sc = read(os.path.join(sourcepath,scinst[:-5],scinst,'*'),starttime=dstarttime,endtime=dendtime)
                # get header from db
                if debug:
                    print ("  Obtained {} data points".format(sc.length()[0]))
                # 2) Remove duplicates
                if debug:
                    print ("  Removing duplicates")
                bef = sc.length()[0]
                sc = sc.removeduplicates()
                aft = sc.length()[0]
                if debug:
                    print ("  -> dropped {} duplicate values".format(bef-aft))
                if sc.length()[0] > 1:
                    if debug:
                        print ("  Getting gaps and data base meta info:")
                    sc = sc.get_gaps()
                    sc.header = db.fields_to_dict(scinst)
                    # flag it
                    if debug:
                        print ("  Flagging with existing flags:")
                        print ("  a) from DB:")
                    flaglist = db.flags_from_db(scinst[:-5], starttime=dstarttime, endtime=dendtime) #data.header['SensorID'])
                    if debug:
                        print ("     Length of DB flaglist: {}".format(len(flaglist)))
                        print ("  b) from file:")
                    if not flagfile:
                        flagfile = os.path.join(outpath,'magpy',scinst[:-5]+'_flags.json')
                    if os.path.isfile(flagfile):
                        fileflaglist = flagging.load(flagfile, begin=dstarttime, end=dendtime)
                        if debug:
                            print("  Loaded {} flags from file {} in the time range {} to {}".format(len(fileflaglist), flagfile, dstarttime, dendtime))
                        if len(flaglist) > 0 and len(fileflaglist) > 0:
                            flaglist = flaglist.join(fileflaglist)
                        elif not len(flaglist) > 0:
                            flaglist = fileflaglist
                    if debug:
                        print("  Found a sum of {} flags".format(len(flaglist)))
                    #sc.header["DataFlags"] = flaglist
                    if flaglist:
                        if debug:
                            print("  Removing flagged data before applying deltas and time shifting")
                        sc = flaglist.apply_flags(sc, mode='drop')
                        sc.header["DataFlags"] = None
                        print ("   -> Done")
                    if debug:
                        print ("  Applying offsets and timeshifts")
                    sc = sc.apply_deltas()
                    if debug:
                        print ("  -> final length: {}".format(sc.length()[0]))
                    sc = sc.removeduplicates()
                    if debug:
                        print ("  -> final length after duplicate removal: {}".format(sc.length()[0]))
                    if not runmode=='firstrun' and not skipsec:
                        hsc = sc.copy()
                        if debug:
                            print ("  Drop all columns except f and comments for high resolution storage")
                        for el in DataStream().KEYLIST:
                            if not el in ['time','f','flag','comment']:
                                hsc = hsc._drop_column(el)
                        #print ("Resample to one second")   # Done in scalarcomb  -> resample deletes flags
                        #hsc = hsc.resample(['f','flag','comment'],period=1)
                        if debug:
                            print ("  Write monthly high resolution file with full info (length {}".format(hsc.length()[0]))
                        fname = '{}_scalar_oc_sec_{}_'.format(scinst,runmode)
                        hsc.write(os.path.join(outpath,'magpy'),filenamebegins=fname,dateformat='%Y%m',coverage='month',mode='replace',format_type='PYCDF')
                    if debug:
                        print ("  Filtering")
                    sc = sc.filter(missingdata='mean')
                    if debug:
                        print ("  Fill gaps with nan")
                    sc = sc.get_gaps()
                    if debug:
                        print ("  Keeping only f column")
                    for el in DataStream().KEYLIST:
                        if not el in ['time','f']:
                            sc = sc._drop_column(el)
                    #mp.plot(sc)
            except:
                pass

            if debug:
                print("  Writing data")
            fminname = '{}_scalar_oc_min_{}_{}'.format(scinst, runmode, year)
            sc.write(os.path.join(outpath, 'magpy'), filenamebegins=fminname, dateformat='%Y', coverage='all', mode='replace', format_type='PYCDF')
            if debug:
                print("Done")
                print("---------------------------")


def variocomb(runmode, config=None, debug=False):
    """
    DESCRIPTION
        create a merged variation file by filling gaps with secondary, teriary etc data sets
    """
    if not config:
        config = {}
        print ("No configuration data existing")
        return

    outpath = config.get('outpath',"/tmp")

    print (" ---------------------------------------------------- ")
    print (" --- Combining Variometer data and merging with F --- ")
    print (" ---------------------------------------------------- ")
    year = config.get('year')
    vainstlist = config.get('vainstlist')

    print ("Dealing with minute data for {}".format(year))
    refmin = DataStream()
    if runmode in ['secondrun','thirdrun']:
        writedefinitive = True
    else:
        writedefinitive = False
    for idx,vainst in enumerate(vainstlist):
        if runmode in ['firstrun']:
            fname = '{}_vario_cbmin_{}.cdf'.format(vainst, year)
        else:
            fname = '{}_vario_blmin_{}_{}.cdf'.format(vainst,runmode,year)
        if idx == 0:
                print ("-------------------------------------")
                print ("Base data: {}".format(vainst))
                print ("-------------------------------------")
                refmin = read(os.path.join(os.path.join(outpath,'magpy',fname)))
                print ("Got {} data points in reference min file".format(refmin.length()[0]))
                refmin = refmin.bc()
                if not refmin.header.get('DataAbsFunctionObject') and runmode == 'firstrun':
                    print ("  Firstrun: baseline has already been applied in simplebaseline adoption")
                print ("Baseline corrected")
                refmin = refmin._drop_column('t1')
                refmin = refmin._drop_column('t2')
                refmin = refmin._drop_column('var1')
                refmin = refmin._drop_column('var2')
                refmin = refmin._drop_column('var5')
                refmin = refmin._drop_column('str1')
                print ("Drop line with nan values")
                refmin = refmin._drop_nans('x')
                refmin = refmin._drop_nans('y')
                refmin = refmin._drop_nans('z')
        else:
                print ("-------------------------------------")
                print ("Adding data from {}".format(vainst))
                print ("-------------------------------------")
                addmin = read(os.path.join(os.path.join(outpath,'magpy',fname)))
                print ("Got {} data points in min file".format(addmin.length()[0]))
                addmin = addmin.bc()
                if not refmin.header.get('DataAbsFunctionObject') and runmode == 'firstrun':
                    print ("  Firstrun: baseline has already been applied in simplebaseline adoption")
                print ("Baseline corrected")
                addmin = addmin._drop_column('t1')
                addmin = addmin._drop_column('t2')
                addmin = addmin._drop_column('var1')
                addmin = addmin._drop_column('var2')
                addmin = addmin._drop_column('var5')
                addmin = addmin._drop_column('str1')
                print ("Drop line with nan values")
                addmin = addmin._drop_nans('x')
                addmin = addmin._drop_nans('y')
                addmin = addmin._drop_nans('z')
                print ("Merging timeseries")
                refmin = join_streams(refmin,addmin)
                print ("Reference now contains {} data points (min)".format(refmin.length()[0]))
    if refmin.length()[0] > 0:
        print ("Minute data combined")
        combminute = True
        refmin = refmin.get_gaps()
        refmin.write(os.path.join(outpath,'magpy'),filenamebegins='CobsV_min_{}'.format(runmode),coverage='all',mode='overwrite',format_type='PYCDF')
    if writedefinitive:
        #read CobsF data if existing
        reff = read(os.path.join(os.path.join(outpath,'magpy','CobsF_min_{}*'.format(runmode))))
        if reff.length()[0] > 0:
            print ("Writing definitive one minute data")
            ts, te = refmin.timerange()
            reff = reff.trim(ts, te+timedelta(seconds=reff.samplingrate()))
            if debug:
                print("Minute V coverage:", refmin.timerange(), refmin.samplingrate())
                print("Minute F coverage", reff.timerange(), reff.samplingrate())
            definitive = merge_streams(refmin,reff, keys=['f'])
            #mp.plot(definitive)
            definitive.write(os.path.join(outpath,'magpy'),filenamebegins='Definitive_min_mag_{}_{}'.format(year,runmode),coverage='all',mode='overwrite',format_type='PYCDF')


def variocorr(runmode, config=None, startmonth=0, endmonth=12, skipsec=False, flagfile=None, basefile=None, debug=False):
    """
    DEFININTION:
        calculate variometer files with baseline and flagging information
    PARAMETER:
        flagfile: a file containing flagging information for the variometer
        basefile: use basevalue data from this file, otherwise use MagPys default paths to find basevalue data
        runmode:  firstrun  -> xxx
                  secondrun ->
                  thidrun   ->
                  adjusted -> use the last two days, read and write from DB
                  quasidefinitive -> use time range since last run put that to config
        skipsec: do not write second data files
    APPLICATION:
        adjusted:  config[

    """
    if not config:
        config = {}
        print ("No configuartion data existing")
        return

    base = config.get('base','')
    db = config.get('primaryDB')
    outpath = config.get('outpath',"/tmp")
    ppier = config.get('primarypier', 'A2')
    blvabb = config.get('blvabb')
    stationid = config.get('obscode','') # TODO cleanup configuration files and use same notoation in all
    if not stationid:
        stationid = config.get('station', '')
    diflagfile= config.get('diflagfile', '')
    vainstlist = config.get('vainstlist',[])
    year = config.get('year')
    testdate = config.get('testdate',None)
    scinstlist = config.get('scinstlist')

    sourcepath = os.path.join(base, 'archive', stationid)
    prf = scinstlist[0] # Required to load the correct basevalue file
    rotangledict = {}

    if debug:
        print ("##########################################################")
        print (" ----------------------------------------------------------------------- ")
        print (" ----------------  Creating minute filtered variodata  ----------------- ")
        print (" -------------------  apply constant/true baseline       -------------------- ")
        print (" ----------------------------------------------------------------------- ")

    def update_rot(stream,year,alpha,beta=None):
            exist = stream.header.get('DataRotationAlpha','')
            if not exist == '':
                val = '{},{}_{}'.format(exist,year,alpha)
            else:
                val = '{}_{}'.format(year,alpha)
            stream.header['DataRotationAlpha'] = val
            if beta:
                exist = stream.header.get('DataRotationBeta','')
                if not exist == '':
                    val = '{},{}_{}'.format(exist,year,beta)
                else:
                    val = '{}_{}'.format(year,beta)
                stream.header['DataRotationBeta'] = val


    if runmode in ['adjusted','quasidefinitive']:
        startmonth = datetime.now(timezone.utc).month-1
        endmonth = datetime.now(timezone.utc).month
        year = datetime.now(timezone.utc).year

    #### If a testdate is provided (single day)
    tdate = None
    if testdate:
        #extract date and month
        tdate = methods.testtime(testdate)
        if not tdate.year == year:
            print ("Year not fitting - ignoring testdate")
            tdate = None
        else:
            startmonth = tdate.month-1
            endmonth = tdate.month

    #### READ MONTHLY ONE SECOND DATA
    for i in range(startmonth,endmonth):   # 0,12
        month = str(i+1).zfill(2)
        nextmonth = str(i+2).zfill(2)
        nextyear = year
        if nextmonth == '13':
            nextyear = year+1
            nextmonth = '01'
        for absidx,inst in enumerate(vainstlist):
            if debug:
                print ("-------------------------------------")
                print ("Running analysis for {} with runmode {}".format(inst,runmode))
                if not runmode in ['adjusted']:
                    print ("  Using Variometer data from path: {}".format(os.path.join(sourcepath, inst[:-5], inst, '*')))

            if runmode in ['adjusted']:
                dendtime = datetime.now(timezone.utc).replace(tzinfo=None)
                dstarttime = dendtime - timedelta(days=2)
            elif runmode in ['quasidefinitive']:
                # TODO get start and enddate from config
                dendtime = datetime.now(timezone.utc).replace(tzinfo=None)
                dstarttime = dendtime - timedelta(days=2)
            elif not tdate: #definitive
                dstarttime=datetime.strptime(str(year)+'-'+month+'-01',"%Y-%m-%d")
                dendtime=datetime.strptime(str(nextyear)+'-'+nextmonth+'-01',"%Y-%m-%d")
            else:
                dstarttime=tdate
                dendtime=tdate+timedelta(days=1)
            dstarttime = dstarttime.replace(tzinfo=None)
            dendtime = dendtime.replace(tzinfo=None)
            if debug:
                print ("  Dealing with data between {} and {}".format(dstarttime,dendtime))

            try:
                if not runmode in ['adjusted']:
                    va = read(os.path.join(sourcepath, inst[:-5], inst, '*'),starttime=dstarttime,endtime=dendtime)
                else:
                    #read DB
                    va = db.read(inst,starttime=dstarttime,endtime=dendtime)
                # raw data
            except:
                va = DataStream()
            if debug:
                print ("  -> got {} datapoints".format(len(va)))
            if len(va) > 1:
                # 1) get header from db
                # ---------------------
                if debug:
                    print ("MEANS: {},{},{}".format(va.mean("x"),va.mean("y"),va.mean("z")))
                dhead = db.fields_to_dict(inst)
                if dhead:
                    va.header = dhead
                if va.header.get('SensorID','') == '':
                    va.header['SensorID'] = inst[:-5]
                # 2) Remove duplicates
                if debug:
                    print ("  Removing duplicates")
                bef = va.length()[0]
                va = va.removeduplicates()
                aft = va.length()[0]
                if debug:
                    print ("  -> dropped {} duplicate values".format(bef-aft))
                # 3) Fill gaps with nan
                if debug:
                    print ("  Filling gaps")
                va = va.get_gaps()
                # 4) flag data
                # ---------------------
                #  --a. add flaglist from db
                if debug:
                    print ("  Flagging")
                vaflagsdropped = False
                vaflaglist = db.flags_from_db(va.header['SensorID'], starttime=dstarttime, endtime=dendtime)
                if debug:
                    print("  Loaded {} flags from DB for this time range".format(len(vaflaglist)))
                #  --b. add flaglist from file
                if not flagfile:
                    flagfile = os.path.join(outpath,'magpy',inst[:-5]+'_flags.json')
                fileflaglist = flagging.load(flagfile, begin=dstarttime, end=dendtime)
                if debug:
                    print("  Loaded {} flags from file in time range {} to {}".format(len(fileflaglist), dstarttime, dendtime))
                if len(vaflaglist) > 0 and len(fileflaglist) > 0:
                    vaflaglist = vaflaglist.join(fileflaglist)
                elif not len(vaflaglist) > 0:
                    vaflaglist = fileflaglist
                if debug:
                    print ("  Found a sum of {} flags".format(len(vaflaglist)))
                    print ("  Applying {} flags".format(len(vaflaglist)))

                va.header["DataFlags"] = vaflaglist # will be kept until second data is saved

                #va = flaglist.apply_flags(va, mode='drop')
                ## Checkpoint
                #mp.plot(va,variables=['x','y','z'],annotate=True)
                if debug:
                    print ("   MEANS: {},{},{}".format(va.mean("x"),va.mean("y"),va.mean("z")))
                # 5) apply compensation
                # ---------------------
                if debug:
                    print("  Applying compensation values")
                va = va.compensation(skipdelta=True)
                if debug:
                    print ("   MEANS: {},{},{}".format(va.mean("x"),va.mean("y"),va.mean("z")))
                # 6) apply delta values
                # ---------------------
                if debug:
                    print("  Applying delta values")
                va = va.apply_deltas()
                # 7) determine and apply rotation
                # ---------------------
                if debug:
                    print("  Rotation values ...")   ### should be done after removal of flags....
                rotstring = va.header.get('DataRotationAlpha','')
                rotdict = string2dict(rotstring,typ='oldlist')
                if debug:
                    print ("  Existing rotation angle alpha for {}: {}".format(str(year), rotdict.get(str(year),'')))
                betastring = va.header.get('DataRotationBeta','')
                betadict = string2dict(betastring,typ='oldlist')
                if debug:
                    print ("  Existing rotation angle beta for {}: {}".format(str(year), betadict.get(str(year),'')))
                if runmode == 'firstrun' or rotdict.get(str(year),0) == 0:
                    print (" Initial run: Rotation value will be determined")
                    print ("   Need to drop flagged data now...")
                    if vaflaglist and not vaflagsdropped:
                        va = vaflaglist.apply_flags(va, mode='drop')
                        va.header['DataFlags'] = None
                        vaflagsdropped = True
                    print ("   Getting rotation angle - please note: correct beta determination requires DI data for reference inclination")
                    alpha, beta, gamma = va.determine_rotationangles(referenceD=0.0, referenceI=None)
                    rotangle = alpha
                    print ("  Determined rotation angle alpha for {}: {}".format(va.header.get('SensorID'),rotangle))
                    print ("  !!! Please note: These  new rotation angles are not yet applied.")
                    print ("  !!! Please note: They are used for secondrun onwards.")
                    print ("  !!! Please note: update DB with correct rotation angles. ")
                else:
                    rotangle = float(rotdict.get(str(year),'0.0'))
                    beta = float(betadict.get(str(year),'0.0'))
                    print("  Applying rotation (after firstrun): Using rotation angles alpha= {} and beta={}".format(rotangle, beta))
                    # Only apply full year values - as baseline calc uses them as well
                    va = va.rotation(alpha=rotangle, beta=beta)
                orgva = va.copy()

                if debug:
                    print ("   MEANS: {},{},{}".format(va.mean("x"),va.mean("y"),va.mean("z")))
                # 8) calc baseline
                # ---------------------
                # requirement: base value data has been cleaned up
                basename = ""
                if debug:
                    print ("Getting basevalue data now (primary pier only):", basefile, inst, prf, ppier)
                if runmode in  ['firstrun','adjusted','quasidefinitive']:
                    scalar = prf[:-5]
                    basename = "{}_{}_{}_{}_{}.txt".format(blvabb, inst[:-5], scalar, ppier, year)
                elif runmode in ['secondrun','thirdrun']:
                    scalar = prf[:-5]
                    basename = "{}_{}_{}_{}_{}_{}.txt".format(blvabb, inst[:-5], scalar, ppier, year, runmode)
                    #basename = 'BLVcomp_'+inst[:-5]+'_'+scalar+'_A2_'+str(year)+'_'+runmode+'.txt'
                if not basefile:
                    basevaluefile = os.path.join(outpath,'magpy',basename)
                else:
                    basevaluefile = basefile

                if debug:
                    print ("  Applying baseline file: {}".format(basevaluefile))
                if os.path.isfile(basevaluefile):
                    absr = read(basevaluefile)
                    print ("  Flagging baseline ...")
                    flaglist = absr.header.get("DataFlags")
                    if flaglist:
                        print("   -> dropping {} flags in baseline file".format(len(flaglist)))
                        absr = flaglist.apply_flags(absr, mode='drop')
                    if debug:
                        print ("  Basevalue SensorID:", absr.header.get('SensorID'))
                        print ("   -- Dropping basedata flags from DB")
                    blvflaglist = db.flags_from_db(absr.header.get('SensorID'))
                    if debug:
                        print ("   -> {} flags".format(len(blvflaglist)))
                    if blvflaglist:
                        absr = blvflaglist.apply_flags(absr, mode='drop')
                    if debug:
                        print ("   -- Dropping basedata flags from file: {}".format(diflagfile))
                    fiflaglist = flagging.load( diflagfile, sensorid=absr.header.get('SensorID'))
                    if debug:
                        print ("   -> {} flags".format(len(fiflaglist)))
                    if fiflaglist:
                        absr = fiflaglist.apply_flags(absr, mode='drop')

                    # Checkpoint
                    #mp.plot(absr)
                    if runmode in ['firstrun','adjusted']:
                        if vaflaglist and not vaflagsdropped:
                            print("  Baseline adoption: dropping flagged data")
                            va = vaflaglist.apply_flags(va, mode='drop')
                            vaflagsdropped = True
                        va = va.get_gaps()
                        print ("  Filtering data")
                        va = va.filter(missingdata='mean')
                        # Checkpoint
                        #mp.plot(va,variables=['x','y','z'])
                        va = va.get_gaps()
                        print ("  Length after gaps removal (min):", va.length()[0])
                        #### GET CONSTANT BASEVALUE
                        print ("  Trimming basevalue file to get mean value")
                        bl = 'cbmin'
                        if str(year) == '2021' and inst == 'LEMI036_2_0001_0002' and i >= 6:
                            absr = absr.trim(starttime=str(year)+'-07-08',endtime=str(year+1)+'-01-01')
                        elif str(year) == '2021' and inst == 'LEMI036_2_0001_0002' and i < 6:
                            print ("....here")
                            absr = absr.trim(starttime=str(year)+'-01-01',endtime=str(year+1)+'-06-30')
                        else:
                            absr = absr.trim(starttime=str(year)+'-01-01',endtime=str(year+1)+'-01-01')
                        print (" -> remaining DI measurements: ", absr.length())
                        absr = absr._drop_nans('dx')
                        bh, bhstd = absr.mean('dx',meanfunction='median',std=True)
                        bd, bdstd = absr.mean('dy',meanfunction='median',std=True)
                        bz, bzstd = absr.mean('dz',meanfunction='median',std=True)

                        print (" Basevalues for {}:".format(year))
                        print (" Delta H = {a} +/- {b}".format(a=bh, b=bhstd))
                        print (" Delta D = {a} +/- {b}".format(a=bd, b=bdstd))
                        print (" Delta Z = {a} +/- {b}".format(a=bz, b=bzstd))

                        print ("  Performing constant basevalue correction")
                        va = va.simplebasevalue2stream([bh,bd,bz])
                    else:
                        bl = 'blmin_{}'.format(runmode)
                        ### TODO the steps below require that basevalues are already calculated with a corrected data set
                        print("Determining baseline")
                        starttime = str(year-1)+'-12-01'
                        endtime = str(year+1)+'-02-01'
                        funclist = []
                        # What about a baseline jump i.e. two or more baseline fits ?
                        # get timeranges and baseline fit parameters from database
                        print (" -> getting baseline fit parameter from database")
                        ett = methods.testtime(endtime)

                        while ett > methods.testtime(starttime):
                            basedict = db.get_baseline(va.header.get('SensorID',''),date=ett)
                            if debug:
                                print ("Got baseline boundary conditions from DB:", basedict)
                            # get the latest (highest ID) input of the basedictionary
                            latestbaseind = max([int(basein) for basein in basedict])
                            baseinput =  basedict.get(latestbaseind, {})
                            bstart = baseinput.get("MinTime", "1777-04-30")
                            bend = baseinput.get("MaxTime", "2223-04-30")
                            if methods.testtime(bstart) < methods.testtime(starttime):
                                 stt = starttime
                            else:
                                 stt = datetime.strftime(methods.testtime(bstart),"%Y-%m-%d")
                            if methods.testtime(bend) > methods.testtime(endtime):
                                 ett = endtime
                            else:
                                 ett = datetime.strftime(methods.testtime(bend),"%Y-%m-%d")
                            fu = baseinput.get("BaseFunction", 'mean')
                            kn = baseinput.get("BaseKnots", '0.3')
                            de = baseinput.get("BaseDegree", '1')
                            print (" => Adding fit with typ {}, knots {}, degree {} between {} and {}".format(fu, float(kn), float(de),stt, ett))
                            try:
                                funclist.append(va.baseline(absr, extradays=0, fitfunc=fu, knotstep=float(kn), fitdegree=float(de),startabs=stt,endabs=ett))
                                print (" => Done")
                            except:
                                print (" => Failed to add baseline parameters")
                            ett = methods.testtime(bstart)-timedelta(days=1)

                    print ("  AbsInfo in stream: {}".format(va.header.get('DataAbsInfo')))
                    if debug:
                        print("   MEANS: {},{},{}".format(va.mean("x"), va.mean("y"), va.mean("z")))

                    if runmode in ['secondrun','thirdrun','adjusted','quasidefinitve'] and not skipsec: # Save high res data when finishing
                        hva = va.copy()
                        if not runmode in ['adjusted']:
                            print ("  Updating rotation information...")
                            update_rot(hva,year,rotangle, beta) ## new 2017
                        #hva = hva.resample(keys=['x','y','z','t1','t2','var2','flag','comment'], period=1)
                        print ("  Write monthly high resolution file with full info")
                        if runmode in ['adjusted','quasidefinitve']:
                            # write to DB (destination)
                            if vaflaglist and not vaflagsdropped:
                                hva = vaflaglist.apply_flags(hva, mode='drop')
                                vaflagsdropped = True
                            hva.header['DataID'] = "{}adjusted_VARIO_0001_0001".format()
                            hva.header['SensorID'] = "{}adjusted_VARIO_0001".format()
                            db.write(hva)
                        else:
                            fname = '{}_vario_sec_{}_'.format(inst,runmode)
                            hva.write(os.path.join(outpath,'magpy'),filenamebegins=fname,dateformat='%Y%m',coverage='month',mode='replace',format_type='PYCDF')
                    print("  remove, gaps and filter:", va.length()[0])
                    if vaflaglist and not vaflagsdropped:
                        va = vaflaglist.apply_flags(va, mode='drop')
                        vaflagsdropped = True
                    #print(va.length())
                    va.header["DataFlags"] = None
                    if not runmode == 'firstrun':
                        va = va.get_gaps()
                        va = va.filter(missingdata='mean')
                    va = va.get_gaps()
                    if debug:
                        print ("  Absolute data has now been loaded and applied - determining true beta now")
                        print ("  Get reference Inclination for the specific month:")
                    redabsr = absr.trim(starttime=dstarttime,endtime=dendtime)
                    meanI = redabsr.mean('x')
                    if debug:
                        print ("  Mean inclination in abs data = {}".format(meanI))
                    alpha, beta, gamma = orgva.determine_rotationangles(referenceI=meanI)

                    if rotangledict.get(inst,[]) == []:
                        rotangledict[inst] = [[alpha,beta]]
                    else:
                        rotangledict[inst].append([alpha,beta])

                    if debug:
                        print ("  Writing")
                    if debug:
                        print("   MEANS: {},{},{}".format(va.mean("x"), va.mean("y"), va.mean("z")))
                    if va.length()[0] > 0:
                        if runmode in ['adjusted','quasidefinitve']:
                            va.header['DataID'] = "{}adjusted_VARIO_0001_0002".format()
                            va.header['SensorID'] = "{}adjusted_VARIO_0001".format()
                            db.write(va)
                        else:
                            va.write(os.path.join(outpath,'magpy'), filenamebegins=inst+'_vario_'+bl+'_'+str(year),dateformat='%Y',mode='replace',coverage='all',format_type='PYCDF')
                else:
                    print (" !!! Basevalue file not existing")


    print ("----------------------------------------------------------")
    print ("Variometer data analyzed and filtered for months {} until first of {}".format(startmonth+1, endmonth+1))
    print ("Runmode: {}".format(runmode))
    print ("Rotation angles: {}".format(rotangledict))
    for inst in vainstlist:
        print ("Extracting rotation for variometer {}".format(inst))
        rotanglelist = rotangledict.get(inst,[])
        if len(rotanglelist) == 2:
            try:
                meanrotangle = np.nanmean(rotanglelist,axis=0)
            except:
                meanrotangle = np.mean(rotanglelist,axis=0)
            print (" -> Average rotation angle for {}: alpha={}, beta={}".format(inst,meanrotangle[0],meanrotangle[1]))
    print ("----------------------------------------------------------")
    print ("##########################################################")
    return rotangledict



class TestDefinitive(unittest.TestCase):
    """
    Test environment for all methods
    """

    def test_variocorr(self):
        runmode ="firstrun"
        config = mm.get_conf(os.path.join('..', 'conf', 'basevalue.cfg'))
        config['dbcredentials'] = "cobsdb"
        startdate, enddate = None, "2025-12-31"
        varios, scalars, piers = ["LEMI036_1_0002_0002","FGE_S0252_0002_0002"],["GP20S3NSS2_012201_0001_0001"],["A2"]
        config = basevalue.check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=True)
        config['blvdatapath'] = "/tmp"
        config['base'] = os.path.abspath("../test")
        config['obscode'] = "WIC"
        config['didatapath'] = os.path.abspath("../test/archive/WIC/DI/analyze")
        config['blvabb'] = "BLV"
        config['testdate'] = "2025-05-14"
        basefile = os.path.join('..', 'test', 'archive', 'WIC', 'DI', 'data', 'BLVcomp_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2.txt')
        rotangledict = variocorr(runmode, config=config, startmonth=0, endmonth=12, skipsec=False, basefile=basefile, debug=True)
        print ("RESULT", rotangledict)
        self.assertTrue(rotangledict)
        us = create_rotation_sql(rotangledict,config=config)
        print ("RESULT", us)
        self.assertTrue(us)

    def test_scalarcorr(self):
        runmode ="firstrun"
        config = mm.get_conf(os.path.join('..', 'conf', 'basevalue.cfg'))
        config['dbcredentials'] = "cobsdb"
        startdate, enddate = None, "2025-12-31"
        varios, scalars, piers = ["LEMI036_1_0002_0002"],["GP20S3NSS2_012201_0001_0001"],["A2"]
        config = basevalue.check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=True)
        config['blvdatapath'] = "/tmp"
        config['base'] = os.path.abspath("../test")
        config['obscode'] = "WIC"
        config['didatapath'] = os.path.abspath("../test/archive/WIC/DI/analyze")
        config['blvabb'] = "BLV"
        config['testdate'] = "2025-05-14"
        scalarcorr(runmode, config=config, startmonth=0, endmonth=12, skipsec=False, debug=True)
        t = False
        if os.path.isfile("/tmp/magpy/GP20S3NSS2_012201_0001_0001_scalar_oc_min_firstrun_2025.cdf"):
            t = True
        self.assertTrue(t)

    def test_variocomb(self):
        runmode ="firstrun"
        config = mm.get_conf(os.path.join('..', 'conf', 'basevalue.cfg'))
        config['dbcredentials'] = "cobsdb"
        startdate, enddate = None, "2025-12-31"
        varios, scalars, piers = ["LEMI036_1_0002_0002","FGE_S0252_0002_0002"],["GP20S3NSS2_012201_0001_0001"],["A2"]
        config = basevalue.check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=True)
        config['obscode'] = "WIC"
        variocomb(runmode, config=config, debug=True)
        t = False
        if os.path.isfile("/tmp/magpy/CobsV_min_firstrun.cdf"):
            t = True
        self.assertTrue(t)

    def test_scalarcomb(self):
        runmode ="firstrun"
        config = mm.get_conf(os.path.join('..', 'conf', 'basevalue.cfg'))
        config['dbcredentials'] = "cobsdb"
        startdate, enddate = None, "2025-12-31"
        varios, scalars, piers = ["LEMI036_1_0002_0002","FGE_S0252_0002_0002"],["GP20S3NSS2_012201_0001_0001"],["A2"]
        config = basevalue.check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=True)
        config['obscode'] = "WIC"
        scalarcomb(runmode, config=config, debug=True)
        t = False
        if os.path.isfile("/tmp/magpy/CobsF_min_firstrun.cdf"):
            t = True
        self.assertTrue(t)

    def test_pier_diff(self):
        runmode ="firstrun"
        config = mm.get_conf(os.path.join('..', 'conf', 'basevalue.cfg'))
        config['dbcredentials'] = "cobsdb"
        startdate, enddate = None, "2025-12-31"
        varios, scalars, piers = ["LEMI036_1_0002_0002","FGE_S0252_0002_0002"],["GP20S3NSS2_012201_0001_0001"],["A2"]
        config = basevalue.check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=True)
        config['obscode'] = "WIC"
        sqllist = pier_diff(runmode, config=config, debug=False)
        print (sqllist)
        #self.assertTrue(sqllist)


if __name__ == "__main__":
    unittest.main(verbosity=2)


'''
Filename:               lemiprotocol
Part of package:        acquisition
Type:                   Part of data acquisition library

PURPOSE:
        This package will initiate LEMI025 / LEMI036 data acquisition and streaming
        and saving of data.

CONTAINS:
        *LemiProtocol:  (Class - twisted.protocols.basic.LineReceiver)
                        Class for handling data acquisition of LEMI variometers.
                        Includes internal class functions: processLemiData
        _timeToArray:   (Func) ... utility function for LemiProtocol.
        h2d:            (Func) ... utility function for LemiProtocol.
                        Convert hexadecimal to decimal.

IMPORTANT:
        - According to data sheet: 300 millseconds are subtracted from each gps time step
        provided GPStime = GPStime_sent - timedelta(microseconds=300000)
        - upcoming year 3000 bug
DEPENDENCIES:
        twisted, autobahn

CALLED BY:
        magpy.bin.acquisition
'''

# ###################################################################
# Import packages
# ###################################################################
import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
import sys
import numpy as np
import os     # binary data saved directly without acs helper method
from datetime import datetime, timedelta, timezone
from twisted.protocols.basic import LineReceiver
from twisted.python import log
from martas.core import methods as mm
import subprocess
from subprocess import check_call


## Lemi protocol (Lemi025 and Lemi036)
## -------------

class LemiProtocol(LineReceiver):
    """
    Protocol to read LEMI (025/036) data
    This protocol defines the individual sensor related read process.
    It is used to dipatch url links containing specific data.
    Sensor specific coding is contained in method "processData".

    Treatment of GPS and NTP time:
    The protocol determines two times: 1) GPS time provided by the data logger
    2) NTP time of the PC when data is received by serial comm.
    LEMI is always recording GPS time a primary. NTP is stored as secondary time.
    Timediff between GPS and NTP is recorded, and a 1000 line median average is send
    within the DataDictionary (DataNTPTimeDelay).
    Besides, a waring message is generated if NTP and GPS times differ by more than 3 seconds.
    For this comparison (and only here) a typical correction (delay) value is applied to NTP.
    e.g. 2.304 sec for LEMI.
    """
    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        print ("Initializing LEMI")
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.delaylist = []  # delaylist contains up to 1000 diffs between gps and ntp
                             # the median of this values is used for ntp timedelay
        self.timedelay = 0.0
        self.ntp_gps_offset = 2.304 # sec - only used for time testing
        self.timethreshold = 3 # secs - waring if timedifference is larger the 3 seconds
        self.buffererrorcnt = 0
        self.compensation = [0.0,0.0,0.0]
        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)
        # PYTHON version
        self.pvers = sys.version_info[0]

        # Debug mode
        debugtest = confdict.get('debug')
        self.debug = False
        if debugtest == 'True':
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))

        # LEMI Specific
        # Using encode() on str that are meant as binaries. Should work on Py2/3
        self.soltag = (self.sensor[0]+self.sensor[4:7])    # Start-of-line-tag
        self.errorcnt = {'gps':'A', 'time':'0', 'buffer':0}
        self.buffer = ''
        self.gpsstate1 = 'A'
        self.gpsstate2 = 'Z'  # Initialize with Z so that current state is send when startet
        self.gpsstatelst = []
        flag = 0
        if self.pvers > 2:
            self.soltag = self.soltag.encode('ascii')    # Start-of-line-tag
            self.buffer = self.buffer.encode('ascii')
            self.gpsstate1 = self.gpsstate1.encode('ascii')
            self.gpsstate2 = self.gpsstate2.encode('ascii')  # Initialize with Z so that current state is send when startet
        print ("Initializing LEMI finished")


    def connectionMade(self):
        log.msg('  -> {} connected.'.format(self.sensor))

    def connectionLost(self, reason):
        log.msg('  -> {} lost.'.format(self.sensor))

    def h2d(self,x):
        '''
        Hexadecimal to decimal (for format LEMIBIN2)
        ... Because the binary for dates is in binary-decimal, not just binary.
        '''

        y = int(x/16)*10 + x%16
        return y

    def initiateRestart(self):
        log.msg('LEMI - Protocol: Cannot fix problem - restarting process')
        log.msg('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
        self.buffer = ''
        self.buffererrorcnt = 0
        if self.pvers > 2:
            self.buffer = self.buffer.encode('ascii')
        print (" ... performing restart now...")
        try:
            # For some reason restart doesn't work?
            print ("Running check_call...")
            check_call(['/etc/init.d/martas', 'restart'])
        except subprocess.CalledProcessError:
            log.msg('LEMI - Protocol: check_call didnt work')
        except:
            log.msg('LEMI - Protocol: check call problem')



    def processLemiData(self, data):
        """Convert raw ADC counts into SI units as per datasheets"""
        if len(data) != 153:
            log.err('LEMI - Protocol: Unable to parse data of length %i' % len(data))

        #print ("Processing data ...")
        """ TIMESHIFT between serial output (and thus NTP time) and GPS timestamp """

        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        date = datetime.strftime(currenttime, "%Y-%m-%d")
        timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
        outtime = datetime.strftime(currenttime, "%H:%M:%S")
        datearray = mm.time_to_array(timestamp)
        date_bin = struct.pack('<6hL',datearray[0]-2000,datearray[1],datearray[2],datearray[3],datearray[4],datearray[5],datearray[6])   ## Added "<" to pack code to get correct length in new machines

        # define pathname for local file storage
        # (default dir plus hostname plus sensor plus year) and create if not existing
        path = os.path.join(self.confdict.get('bufferdirectory'), self.sensor)

        if not os.path.exists(path):
            os.makedirs(path)

        packcode = "<4cb6B8hb30f3BcBcc5hL"
        header = "LemiBin %s %s %s %s %s %s %d\n" % (self.sensor, '[x,y,z,t1,t2]', '[X,Y,Z,T_sensor,T_elec]', '[nT,nT,nT,deg_C,deg_C]', '[0.001,0.001,0.001,100,100]', packcode, struct.calcsize(packcode))
        sendpackcode = '6hLffflll'
        #headforsend = "# MagPyBin {} {} {} {} {} {} {}".format(self.sensor, '[x,y,z,t1,t2,var2,str1]', '[X,Y,Z,T_sensor,T_elec,VDD,GPS]', '[nT,nT,nT,deg_C,deg_C,V,Status]', '[0.001,0.001,0.001,100,100,10]', sendpackcode, struct.calcsize('<'+sendpackcode))
        headforsend = "# MagPyBin {} {} {} {} {} {} {}".format(self.sensor, '[x,y,z,t1,t2,var2]', '[X,Y,Z,T_sensor,T_elec,VDD]', '[nT,nT,nT,deg_C,deg_C,V]', '[0.001,0.001,0.001,100,100,10]', sendpackcode, struct.calcsize('<'+sendpackcode))
        if self.pvers > 2:
            header = header.encode('ascii')

        # save binary raw data to buffer file ### please note that this file always contains GPS readings
        lemipath = os.path.join(path,self.sensor+'_'+date+".bin")
        if not os.path.exists(lemipath):
            with open(lemipath, "ab") as myfile:
                myfile.write(header)
        try:
            with open(lemipath, "ab") as myfile:
                myfile.write(data+date_bin)
            pass
        except:
            log.err('LEMI - Protocol: Could not write data to file.')

        # unpack data and extract time and first field values
        # This data is streamed via mqtt
        try:
            data_array = struct.unpack("<4cB6B8hb30f3BcB", data)
        except:
            log.err("LEMI - Protocol: Bit error while reading.")

        try:
            newtime = []
            biasx = float(data_array[16])/400.
            biasy = float(data_array[17])/400.
            biasz = float(data_array[18])/400.
            x = (data_array[20])*1000.
            xarray = [elem * 1000. for elem in data_array[20:50:3]]
            y = (data_array[21])*1000.
            yarray = [elem * 1000. for elem in data_array[21:50:3]]
            z = (data_array[22])*1000.
            zarray = [elem * 1000. for elem in data_array[22:50:3]]
            temp_sensor = data_array[11]/100.
            temp_el = data_array[12]/100.
            vdd = float(data_array[52])/10.
            gpsstat = data_array[53]
            gpstime = datetime(2000+self.h2d(data_array[5]),self.h2d(data_array[6]),self.h2d(data_array[7]),self.h2d(data_array[8]),self.h2d(data_array[9]),self.h2d(data_array[10]))-timedelta(microseconds=300000)
            #gps_time = datetime.strftime(gps_array, "%Y-%m-%d %H:%M:%S")
            self.compensation[0] = biasx
            self.compensation[1] = biasy
            self.compensation[2] = biasz
        except:
            log.err("LEMI - Protocol: Number conversion error.")

        if self.pvers > 2:
            gpsstat = gpsstat.decode('ascii')

        #print ("HERE2", packcode, struct.calcsize(packcode))
        processerror = False
        if not gpsstat in ['P','A']:
            print (" ERROR in BINDATA:", struct.unpack("<4cB6B8hb30f3BcB", data))
            print (" Rawdata looks like:", data)
            self.buffererrorcnt += 1
            processerror = True
            if self.buffererrorcnt == 10:
                self.initiateRestart()

        # get the most frequent gpsstate of the last 10 min
        # this avoids error messages for singular one sec state changes
        self.gpsstatelst.append(gpsstat)
        self.gpsstatelst = self.gpsstatelst[-600:]
        self.gpsstate1 = max(set(self.gpsstatelst),key=self.gpsstatelst.count)
        if not self.gpsstate1 == self.gpsstate2:
            log.msg('LEMI - Protocol: GPSSTATE changed to %s .'  % gpsstat)
        self.gpsstate2 = self.gpsstate1

        try:
            # Analyze time difference between GPS and NTP
            timelist = sorted([gpstime,currenttime])
            timediff = timelist[1]-timelist[0]
            delta = timediff.total_seconds()
            if not delta in [0.0, np.nan, None]:
                self.delaylist.append(timediff.total_seconds())
                self.delaylist = self.delaylist[-1000:]
            if len(self.delaylist) > 100:
                try:
                    self.timedelay = np.median(np.asarray(self.delaylist))
                except:
                    self.timedelay = 0.0
            if delta-self.ntp_gps_offset > self.timethreshold:
                self.errorcnt['time'] +=1
                if self.errorcnt.get('time') < 2:
                    log.msg("  -> {} protocol: large time difference observed for {}: {} sec".format(self.sensordict.get('protocol'), self.sensor, delta))
            else:
                self.errorcnt['time'] = 0
        except:
            pass

        ### NNOOOO, always send GPS time - but provide median time delays with the dictionary
        ### check LEMI Records whether secondary time (NTP) is readable and extractable

        # Create a dataarray
        linelst = []
        for idx,el in enumerate(xarray):
            datalst = []
            tincr = idx/10.
            timear = gpstime+timedelta(seconds=tincr)
            gps_time = datetime.strftime(timear.replace(tzinfo=None), "%Y-%m-%d %H:%M:%S.%f")
            datalst = mm.time_to_array(gps_time)
            datalst.append(xarray[idx]/1000.)
            datalst.append(yarray[idx]/1000.)
            datalst.append(zarray[idx]/1000.)
            datalst.append(int(temp_sensor*100))
            datalst.append(int(temp_el*100))
            datalst.append(int(vdd*10))
            ### TODO Add GPS and secondary time to this list
            #datalst.append(gpsstat)
            #current_time = datetime.strftime(currenttime.replace(tzinfo=None), "%Y-%m-%d %H:%M:%S.%f")
            #datalst.extend(current_time)
            linestr = ','.join(list(map(str,datalst)))
            linelst.append(linestr)
        dataarray = ';'.join(linelst)


        if processerror:
            print ("Processing unsuccessful")
            dataarray = ''

        return dataarray, headforsend


    def dataReceived(self, data):
        #print ("Lemi data here!", self.buffer)
        """
        Sometime the code is starting wrongly -> 148 bit length
        self healing not working - Check data
        """

        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')

        flag = 0
        WSflag = 0
        #debug = self.debug

        """
        # Test range
        self.buffer = self.buffer + data
        if not (self.buffer).startswith(self.soltag):
            lemisearch = (self.buffer).find(self.soltag, 6)
            if not lemisearch == -1:
                print ("Lemisearch", lemisearch, self.buffer)
                self.buffer = self.buffer[lemisearch:len(self.buffer)]
        if len(self.buffer) == 153:
            # Process data
            print (self.buffer)
            self.buffer = ''
        """

        try:
            # 1. Found correct data length
            if (self.buffer).startswith(self.soltag) and len(self.buffer) == 153:
                currdata = self.buffer
                self.buffer = ''
                if self.pvers > 2:
                    self.buffer = self.buffer.encode('ascii')
                dataarray, head = self.processLemiData(currdata)
                comp = ''
                if self.pvers > 2:
                    comp = comp.encode('ascii')
                if not dataarray == comp:
                    WSflag = 2

            # 2. Found incorrect data length

            ### Note: this code for fixing data is more complex than the POS fix code
            ### due to the LEMI device having a start code rather than an EOL code.
            ### It can handle and deal with multiple errors:
            ###  - multiple data parts arriving at once
            ###  - databits being lost. Bad string is then deleted.
            ###  - bad bits infiltrating the data. Bad string is deleted.

            if len(self.buffer) > 153:
                if self.debug:
                    log.msg('LEMI - Protocol: Warning: Bufferlength ({}) exceeds 153 characters, fixing...'.format(len(self.buffer)))
                    log.msg("BUFFER: {}".format(self.buffer))
                lemisearch = (self.buffer).find(self.soltag)
                #print '1', lemisearch
                if (self.buffer).startswith(self.soltag):
                    datatest = len(self.buffer)%153
                    dataparts = int(len(self.buffer)/153)
                    log.msg('LEMI - Protocol: datatest {} and dataparts {}'.format(datatest,dataparts))
                    if datatest == 0:
                        if self.debug:
                            log.msg('LEMI - Protocol: It appears multiple parts came in at once, amount N of parts: {}'.format(dataparts))
                        for i in range(dataparts):
                            split_data_string = self.buffer[0:153]
                            if (split_data_string).startswith(self.soltag):
                                if self.debug:
                                    log.msg('LEMI - Protocol: Processing data part N {} in string...'.format(str(i+1)))
                                dataarray, head = self.processLemiData(split_data_string)
                                WSflag = 2
                                self.buffer = self.buffer[153:len(self.buffer)]
                            else:
                                flag = 1
                                break
                    else:
                        for i in range(dataparts):
                            lemisearch = (self.buffer).find(self.soltag, 6)
                            if lemisearch >= 153:
                                split_data_string = self.buffer[0:153]
                                dataarray, head = self.processLemiData(split_data_string)
                                WSflag = 2
                                self.buffer = self.buffer[153:len(self.buffer)]
                            elif lemisearch == -1:
                                log.msg('LEMI - Protocol: No header found. Deleting buffer.')
                                self.buffer = ''
                                if self.pvers > 2:
                                    self.buffer = self.buffer.encode('ascii')
                            else:
                                log.msg('LEMI - Protocol: String contains bad data ({} bits). Deleting. Lemisearchpos: {}'.format(len(self.buffer[:lemisearch]), lemisearch))
                                #self.buffer = ''
                                self.buffer = self.buffer[lemisearch:len(self.buffer)]
                                flag = 1
                                self.buffererrorcnt += 1
                                #break #uncommented on 2018-09-11
                    if self.buffererrorcnt == 10:
                        self.initiateRestart()

                else:
                    log.msg('LEMI - Protocol: Incorrect header. Attempting to fix buffer... Bufferlength: {}'.format(len(self.buffer)))
                    lemisearch = (self.buffer).find(self.soltag, 6)
                    #lemisearch = repr(self.buffer).find(self.soltag)
                    if lemisearch == -1:
                        log.msg('LEMI - Protocol: No header found. Deleting buffer.')
                        self.buffer = ''
                        if self.pvers > 2:
                            self.buffer = self.buffer.encode('ascii')
                    else:
                        self.buffer = self.buffer[lemisearch:len(self.buffer)]
                        log.msg('LEMI - Protocol: Bad data ({} bits) deleted. New bufferlength: {}'.format(lemisearch,len(self.buffer)))
                        flag = 1

            if flag in [0,1]: #flag == 0: #uncommented 2018-09-11
                self.buffer = self.buffer + data

        except:
            log.msg('LEMI - Protocol: Error while parsing data.')
            #Emtpying buffer
            self.buffer = ''
            if self.pvers > 2:
                self.buffer = self.buffer.encode('ascii')
            self.buffererrorcnt += 1
            if self.buffererrorcnt == 10:
                self.initiateRestart()

        ## publish events to all clients subscribed to topic
        if WSflag == 2:

            self.buffererrorcnt = 0
            senddata = False
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(dataarray)
                    self.datacnt += 1
                else:
                    senddata = True
                    dataarray = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", dataarray, qos=self.qos)
                if self.count == 0:
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDescription:{},DataTimeProtocol:{},DataNTPTimeDelay:{},DataCompensationX:{},DataCompensationY:{},DataCompensationZ:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc','').rstrip(),self.sensordict.get('ptime',''),self.timedelay, self.compensation[0],self.compensation[1],self.compensation[2] )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)
                self.count += 1
                if self.count >= self.metacnt:
                    self.count = 0



from __future__ import print_function
from __future__ import absolute_import

# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
import numpy as np
from datetime import datetime, timezone
from twisted.protocols.basic import LineReceiver
from twisted.python import log
from martas.core import methods as mm
import sys


## POS1 protocol
## -------------

class POS1Protocol(LineReceiver):
    """
    Protocol to read POS1 data
    This protocol defines the individual sensor related read process. 
    It is used to dipatch url links containing specific data.
    Sensor specific coding is contained in method "processData".

    Treatment of GPS and NTP time:
    The protocol determines two times: 1) GPS time provided by the data logger
    2) NTP time of the PC when data is received by serial comm.
    Primary time is selected in sensor.cfg (either NTP or GPS)
    Timediff between GPS and NTP is recorded, and a 1000 line median average is send
    within the DataDictionary (DataNTPTimeDelay).
    Besides, a waring message is generated if NTP and GPS times differ by more than 3 seconds.
    For this comparison (and only here) a typical correction (delay) value is applied to NTP.
    e.g. 6.2 sec for POS1.

    """
    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        print ("Begin Initialization of POS1")
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        #log.msg("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.errorcnt = {'time':0}
        self.delaylist = []  # delaylist contains up to 1000 diffs between gps and ntp (with offset)
                             # the median of this values is used for ntp timedelay
        self.timedelay = 0.0
        self.ntp_gps_offset = 6.2 # sec - is only used for time diff checks - true ntp time is stored
        self.timethreshold = 3 # secs - waring if timedifference is larger the 3 seconds

        delimiter = '\x00'
        self.buffer = ''

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)
        print ("  -> End of POS1 initialization")


    def connectionMade(self):
        log.msg('  -> {} connected.'.format(self.sensor))

    def connectionLost(self, reason):
        log.msg('  -> {} lost.'.format(self.sensor))

    def processPos1Data(self, data):
        """Convert raw ADC counts into SI units as per datasheets"""
        if len(data) != 44:
            log.err('POS1 - Protocol: Unable to parse data of length %i' % len(data))

        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(currenttime, "%Y-%m-%d")
        actualtime = datetime.strftime(currenttime, "%Y-%m-%dT%H:%M:%S.%f")
        outtime = datetime.strftime(currenttime, "%H:%M:%S")
        timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
        filename = outdate
        sensorid = self.sensor
        data_bin = None
        gps_time = ""
        datearray = []

        packcode = '6hLLLh6hL'
        header = "# MagPyBin %s %s %s %s %s %s %d" % (self.sensor, '[f,df,var1,sectime]', '[f,df,var1,GPStime]', '[nT,nT,none,none]', '[1000,1000,1,1]', packcode, struct.calcsize('<'+packcode))

        try:
            # Extract data
            data_array = data.split()
            intensity = float(data_array[0])/1000.
            sigma_int = float(data_array[2])/1000.
            err_code = int(data_array[3].strip('[').strip(']'))
            dataelements = datetime.strptime(data_array[4],"%m-%d-%y")
            newdate = datetime.strftime(dataelements,"%Y-%m-%d")
            gps_time = newdate + ' ' + str(data_array[5])[:11]
        except:
            log.err('POS1 - Protocol: Data formatting error.')
            intensity = 0.0
            sigma_int = 0.0
            err_code = 0.0

        try:
            # Analyze time difference between POS1 internal time and utc from PC
            # Please note that the time difference between POS1-GPS (data recorded) 
            # and NTP (data received at PC) can be very large
            # for our POS1 it is 6.2 seconds

            gpstime = datetime.strptime(gps_time, "%Y-%m-%d %H:%M:%S.%f")
            timelist = sorted([gpstime,currenttime])
            timediff = timelist[1]-timelist[0]
            delta = timediff.total_seconds()
            if not delta in [0.0, np.nan, None]:
                self.delaylist.append(delta)
                self.delaylist = self.delaylist[-1000:]
            if len(self.delaylist) > 100:
                try:
                    self.timedelay = np.median(np.asarray(self.delaylist))
                except:
                    self.timedelay = 0.0
            if delta-self.ntp_gps_offset > self.timethreshold:
                self.errorcnt['time'] +=1
                if self.errorcnt.get('time') < 2:
                    log.msg("{} protocol: large time difference observed for {}: {} sec".format(self.sensordict.get('protocol'), sensorid, delta))
            else:
                self.errorcnt['time'] = 0 
        except:
            pass

        if self.sensordict.get('ptime','') in ['NTP','ntp']:
            secondtime = gps_time
            maintime = timestamp
        else:
            maintime = gps_time
            secondtime = timestamp

        try:
            # extract time data
            datearray = mm.time_to_array(maintime)
            sectarray = mm.time_to_array(secondtime)
            try:
                datearray.append(int(intensity*1000))
                datearray.append(int(sigma_int*1000))
                datearray.append(err_code)
                datearray.extend(sectarray)
                data_bin = struct.pack('<'+packcode,datearray[0],datearray[1],datearray[2],datearray[3],datearray[4],datearray[5],datearray[6],datearray[7],datearray[8],datearray[9],datearray[10],datearray[11],datearray[12],datearray[13],datearray[14],datearray[15],datearray[16])
            except:
                log.msg('POS1 - Protocol: Error while packing binary data')
                pass
            if not self.confdict.get('bufferdirectory','') == '':
                mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
        except:
            log.msg('POS1 - Protocol: Error with binary save routine')
            pass

        return ','.join(list(map(str,datearray))), header

    def dataReceived(self, data):
        dataarray = ""
        head = ""
        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        # extract only ascii characters
        if sys.version_info >= (3,0):
            data = data.decode()

        ok = False
        try:
            if len(self.buffer) == 44:
                dataarray, head = self.processPos1Data(self.buffer[:44])
                self.buffer = ''
                try:
                    value = float(dataarray.split(',')[7])
                except:
                    value = 0.0
                if value > 0:
                    ok = True
                else:
                    log.err('POS1 - Protocol: Zero value, skipping. (Value still written to file.)')
            self.buffer = self.buffer + data
            if len(self.buffer) > 44:
                log.msg('POS1 - Protocol: Warning: Bufferlength (%s) exceeds 44 characters, fixing...' % len(self.buffer))
                if repr(data).endswith("x00'"):    # check if last part read is end of POS string.
                    datatest = (len(self.buffer))%44
                    # OPTION 1: Data is good, but too much arrived at once. Split and process.
                    if datatest == 0:
                        dataparts = int(len(self.buffer)/44)
                        log.msg('POS1 - Protocol: It appears multiple parts came in at once, # of parts:', dataparts)
                        for i in range(dataparts):
                            split_data_string = self.buffer[i*44:(i*44)+44]
                            log.msg('POS1 - Protocol: Processing data part # %s in string (%s)' % (str(i+1), split_data_string))
                            dataarray, head = self.processPos1Data(split_data_string)
                            try:
                                value = float(dataarray.split(',')[7])
                            except:
                                value = 0.0
                            if value > 0:
                                ok = True
                            else:
                                log.err('POS1 - Protocol: Zero value, skipping. (Value still written to file.)')
                        self.buffer = ''
                    # OPTION 2: Data is bad; bit was lost.
                    else:
                        log.msg('POS1 - Protocol: String contains bad data. Deleting. (String content: %s)' % self.buffer)
                        self.buffer = ''              # If true, bad data. Log and delete.
                else:    # if last part read is not end of POS string, continue reading.
                    log.msg('POS1 - Protocol: Attempting to fix buffer... last part read:', self.buffer[-10:], "Bufferlength:", len(self.buffer))
        except:
            print('{}: Data seems not be POS1Data: Looks like {}'.format(self.sensordict.get('protocol'),data))
            ok = False


        if ok:
            senddata = False
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(dataarray)
                    self.datacnt += 1
                else:
                    senddata = True
                    dataarray = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", dataarray, qos=self.qos)
                if self.count == 0:
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{},DataNTPTimeDelay:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime',''), self.timedelay )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)

                self.count += 1
                if self.count >= self.metacnt:
                    self.count = 0




#!/usr/bin/env python

from matplotlib.dates import date2num, num2date
from datetime import datetime, timezone
import numpy as np
import time

"""
init file with typus, distances, averaging parameters (i.e. 30), frequnecy and current

Alternatively, get commands from sensors.cfg just like for arduino
frequency and current would work with standard init
main routine reads init and then runs in passive mode
"""

"""
port = '/dev/ttyUSB0'
baudrate='19200'
eol = '\r'
commandlist = ["w","O","c","o","M"]
currdic = {"m":"1uA","n":"10uA","o":"100uA","p":"1mA","q":"5mA","r":"15mA","s":"50mA","t":"100mA"}
freqdic = {"a":"0.26Hz","b":"0.52Hz","c":"1.04Hz","d":"2.08Hz","e":"4.16Hz","f":"8.33Hz","g":"12.5Hz","h":"25Hz"}

commanddict = {"v":"Softwareversion","w":"external voltage","M":"single measurement","N":"Transmitter output voltage","O":"Battery Voltage","P":"Self potential","Q":"Cont measure start","R":"Cont measure stop","S":"Transmitter on","T":"Transmitter Off","W":"External electrodes on","X":"External electrodes off"}

def lineread(ser,eol):
            # FUNCTION 'LINEREAD'
            # Does the same as readline(), but does not require a standard
            # linebreak character ('\r' in hex) to know when a line ends.
            # Variable 'eol' determines the end-of-line char: '\x00'
            # for the POS-1 magnetometer, '\r' for the envir. sensor.
            # (Note: required for POS-1 because readline() cannot detect
            # a linebreak and reads a never-ending line.)
            ser_str = ''
            timeout = time.time()+2
            while True:
                char = ser.read().decode()
                if char == eol:
                    break
                if time.time() > timeout:
                    break
                ser_str += char
            return ser_str

def send_command(ser,command,eol,hex=False):

    command = command
    sendtime = date2num(datetime.now(timezone.utc).replace(tzinfo=None))
    ser.write(command.encode())
    response = lineread(ser,eol)
    receivetime = date2num(datetime.now(timezone.utc).replace(tzinfo=None))
    meantime = np.mean([receivetime,sendtime])

    return response, num2date(meantime).replace(tzinfo=None)

def rho(U,I,A,L=0,typus="wenner"):
    rh = 0.0
    if typus in ["schlumberger","Schlumberger","schlum"]:
        rh = np.pi * (((L/2)**2 - (A/2)**2) / A) * U / I
    elif typus in ["half-schlumberger","Half-Schlumberger","half-schlum"]:
        rh = 2 * np.pi * (((L/2)**2 - (A/2)**2) / A) * U / I
    elif typus in ["dipole-dipole","Dipol-Dipol","Dipole-Dipole","dipol-dipol"]:
        n = L/A
        rh = np.pi * n(n+1)(n+2) * A * U / I
    else:
        rh = 2* np.pi * A * U / I
    return rh

def phase(U0,U90):
    # return phase in mrad
    return U90/U0*1000.


ser = serial.Serial(port, baudrate=baudrate , parity='N', bytesize=8, stopbits=1, timeout=2)

A=0.65

test = 1
if test==2:
    commandlist = ["v","w","S","N","M","T","O"]
    for comm in commandlist:
        time.sleep(0.5)
        answer, actime = send_command(ser,comm,eol)
        print("Receiving {}: {}".format(commanddict.get(comm),answer))

if test==1:
    print ("Setting parameters:")
    print ("-------------------")
    meascom = ["v","w","O","c","o"]
    I = 0
    for comm in meascom:
        time.sleep(0.5)
        answer, actime = send_command(ser,comm,eol)
        print("Receiving {}: {}".format(commanddict.get(comm,"Not in list"),answer))
        if comm in ["m","n","o","p","q","r","s","t"]:
            Ist = currdic.get(comm)
            print (Ist)
            if Ist.find("mA") >0:
                I = float(Ist.replace("mA",""))/1000
            else:
                I = float(Ist.replace("uA",""))/1000000

    print (I)
    print ("Running measurement:")
    print ("-------------------")

    for n in range(0,1440):
        # 10 seconds break
        time.sleep(10)
        # start transmission
        comms = ["S","M","T"]
        sum0,sum90 = [],[]
        print ("------------------------")
        print ("Date:", datetime.now(timezone.utc).replace(tzinfo=None))
        for comm in comms:
            #answer,actime = send_command(ser,comm,eol)
            if comm == "M":
                for o in range(1,30):
                    answer,actime = send_command(ser,comm,eol)
                    answer = answer.replace("M","")
                    #print ("Current {}: U0[mV]={},U90[mV]={}".format(currentdic[curr],answer.split(" ")[0],answer.split(" ")[1]))
                    #print ("U0[mV]={},U90[mV]={}".format(answer.split(" ")[0],answer.split(" ")[1]))
                    u0 = float(answer.split(" ")[0])
                    u90 = float(answer.split(" ")[1])
                    sum0.append(u0)
                    sum90.append(u90)
                    if o == 29:
                        print ("U0[mV]={:.3f}, std={:.3f}, percental error={:.2f}%".format(np.mean(sum0),np.std(sum0),np.std(sum0)/np.mean(sum0)/np.sqrt(29)*100.))
                        print ("U90[mV]={:.3f}, std={:.3f}, percental error={:.2f}%".format(np.mean(sum90),np.std(sum90),np.std(sum90)/np.mean(sum90)/np.sqrt(29)*100. ) )
                        print ("Rho[Ohm*m]={:.1f},Phase[mrad]={:.2f}".format(rho(np.mean(sum0)/1000.,I,A),phase(np.mean(sum0),np.mean(sum90))))
                    time.sleep(0.2)
            else:
                answer,actime = send_command(ser,comm,eol)
                print("Receiving {}: {}".format(commanddict.get(comm),answer))

"""

# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
from datetime import datetime, timedelta
from twisted.python import log
from martas.core import methods as mm
import serial


def lineread(ser,eol):
            # FUNCTION 'LINEREAD'
            # Does the same as readline(), but does not require a standard
            # linebreak character ('\r' in hex) to know when a line ends.
            # Variable 'eol' determines the end-of-line char: '\x00'
            # for the POS-1 magnetometer, '\r' for the envir. sensor.
            # (Note: required for POS-1 because readline() cannot detect
            # a linebreak and reads a never-ending line.)
            ser_str = ''
            timeout = time.time()+2
            while True:
                char = ser.read().decode()
                if char == eol:
                    break
                if time.time() > timeout:
                    break
                ser_str += char
            return ser_str


def send_command(ser,command,eol,hex=False):

    command = command
    sendtime = date2num(datetime.now(timezone.utc).replace(tzinfo=None))
    ser.write(command.encode())
    response = lineread(ser,eol)
    receivetime = date2num(datetime.now(timezone.utc).replace(tzinfo=None))
    meantime = np.mean([receivetime,sendtime])

    return response, num2date(meantime).replace(tzinfo=None)


def datetime2array(t):
    return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

## Arduino active request protocol
## --------------------

class FourPLProtocol(object):
    """
    Protocol to read Lippmann 4PL

    Active protocol

    Example sensor.cfg
    4PL_0009009195_0001,USB0,19200,8,1,N,active,None,60,1,FourPL,4PL,0009009195,0001,...

    """

    ## need a reference to our WS-MCU gateway factory to dispatch PubSub events
    ##
    def __init__(self, client, sensordict, confdict):
        self.client = client #self.wsMcuFactory = wsMcuFactory
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensorname = sensordict.get('name')
        self.sensorid = sensordict.get('sensorid')
        # serial
        self.baudrate=int(sensordict.get('baudrate'))
        self.port = confdict['serialport']+sensordict.get('port')
        self.parity=sensordict.get('parity')
        self.bytesize=sensordict.get('bytesize')
        self.stopbits=sensordict.get('stopbits')
        self.timeout=2

        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        self.datalst = []
        self.datacnt = 0
        # Sampling rate specific data
        rate = sensordict.get('rate','0')
        self.metacnt = 1
        try:
            self.rate = int(rate)
        except:
            self.rate = 10 #(a minimum of 10 secs is used)
        if self.rate < 10:
            log.msg('4PLprotocol: Minimum sampling period for 4PL in MARTAS is 10 sec. Adjusting rate...')
            self.rate = 10 # send header at reduced rate
            self.metacnt = 10 # send header at reduced rate
        self.N = int(self.rate/2.)
        log.msg('4PLprotocol: Recording {} individual measurements for mean resistivity.'.format(self.N))
        self.counter = {}

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)

        # switch on debug mode
        debugtest = confdict.get('debug')
        self.debug = False
        if debugtest == 'True':
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))

        # 4PL specific
        #self.sensorid = ''
        self.serialnum = ''
        self.currdic = {"m":"1uA","n":"10uA","o":"100uA","p":"1mA","q":"5mA","r":"15mA","s":"50mA","t":"100mA"}
        self.freqdic = {"a":"0.26Hz","b":"0.52Hz","c":"1.04Hz","d":"2.08Hz","e":"4.16Hz","f":"8.33Hz","g":"12.5Hz","h":"25Hz"}
        self.commanddict = {"v":"Softwareversion","w":"external voltage","M":"single measurement","N":"Transmitter output voltage","O":"Battery Voltage","P":"Self potential","Q":"Cont measure start","R":"Cont measure stop","S":"Transmitter on","T":"Transmitter Off","W":"External electrodes on","X":"External electrodes off"}
        # Specific commands for ultrasonic wind sensor
        # self.addressanemo = '01'
        #self.commands = [{'data':'11TR00005'},{'meta':''},{'head':''}] # ,self.addressanemo+'TR00002']
        #self.commands = [{'data':'01TR00002','c1':'01SH','c2':'01SL'}]
        self.eol = '\r'

        # Commands (take them sensordesc)
        # i.e. sensordesc : owT-swD
        commandlist = sensordict.get('sensordesc').strip().replace('\n','').split('-')
        print ("Length",len(commandlist))
        if not 'arduino sensors' in commandlist and len(commandlist)==5:
            self.commands = {}
            self.commands["mode"] = commandlist[0]
            self.commands["A"] = commandlist[1]
            self.commands["L"] = commandlist[2]
            self.commands["frequency"] = commandlist[3]
            self.commands["current"] = commandlist[4]
            print ("Commands from sensordesc:", self.commands)
        else:
            self.commands = {'mode':'wenner','A':'0.65','L':'0','frequency':'c','current':'o'}
            print ("Default commands selected")

        # call "setcurrandfrequ", requires curret and frequency from commands
        self.A = 0.65 # = float(self.command.get("A")) get from sensorsconf
        self.L = None # = float(self.command.get("L"))get from sensorsconf
        #self.N = 30 # get from sensorsconf (samplingrate must be twice as high) (i.e. take sampling rate and use half of it)
        self.ser = self.connectserial()
        self.I,self.F = self.setcurrandfrequ("c","o") # = self.setcurrandfrequ(self.command.get("frequency"),self.command.get("current"))

    def connectserial(self):
        ser = serial.Serial(self.port, baudrate=int(self.baudrate), parity=self.parity, bytesize=int(self.bytesize), stopbits=int(self.stopbits), timeout=int(self.timeout))
        if not ser.isOpen():
            ser.open()
        return ser

    def setcurrandfrequ(self,freq,curr):
        """
        Obtain frequency and current from code
        """
        I,F = 0,0
        ser = self.ser
        ok = True
        if ok:
            log.msg('4PLprotocol: Initializing frequency and current')
            meascom = ["v","w","O",freq,curr]
            for comm in meascom:
                answer, actime = send_command(ser,comm,self.eol)
                if self.commanddict.get(comm,""):
                    log.msg("Receiving {}: {}".format(self.commanddict.get(comm,""),answer))
                if comm in ["a","b","c","d","e","f","g","h"]:
                    Fst = self.freqdic.get(comm)
                    log.msg("Setting frequency: {}".format(Fst))
                    F = float(Fst.replace("Hz",""))
                if comm in ["m","n","o","p","q","r","s","t"]:
                    Ist = self.currdic.get(comm)
                    log.msg("Setting current: {}".format(Ist))
                    if Ist.find("mA") >0:
                        I = float(Ist.replace("mA",""))/1000
                    else:
                        I = float(Ist.replace("uA",""))/1000000
                time.sleep(0.5)
            log.msg('4PLprotocol: Initializing complete')

        return I,F

    def rho(self,U,I,A,L=0,typus="wenner"):
        """
        """
        rh = 0.0
        if typus in ["schlumberger","Schlumberger","schlum"]:
            rh = np.pi * (((L/2)**2 - (A/2)**2) / A) * U / I
        elif typus in ["half-schlumberger","Half-Schlumberger","half-schlum"]:
            rh = 2 * np.pi * (((L/2)**2 - (A/2)**2) / A) * U / I
        elif typus in ["dipole-dipole","Dipol-Dipol","Dipole-Dipole","dipol-dipol"]:
            n = L/A
            rh = np.pi * n(n+1)(n+2) * A * U / I
        else:
            rh = 2* np.pi * A * U / I
        return rh

    def phase(self,U0,U90):
        # return phase in mrad
        return U90/U0*1000.

    def error(self,sigma,mean,N):
        # return phase in mrad
        return sigma/np.sqrt(N)/mean*100.

    def gettime(self,st,et):
        # calculate mean time and eventually round to minute or hour

        def roundTime(dt=None, roundTo=60):
            """
            Round a datetime object to any time lapse in seconds
            dt : datetime.datetime object, default now.
            roundTo : Closest number of seconds to round to, default 1 minute.
            Author: Thierry Husson 2012
            """
            if dt == None:
                dt = datetime.now(timezone.utc).replace(tzinfo=None)
            seconds = (dt.replace(tzinfo=None) - dt.min).seconds
            rounding = (seconds+roundTo/2) // roundTo * roundTo
            return dt + timedelta(0,rounding-seconds,-dt.microsecond)

        dmean = num2date(np.mean([date2num(st),date2num(et)]))
        actime = None
        if self.rate >= 60:
            actime = roundTime(dmean,roundTo=60)
        elif self.rate >= 3600:
            actime = roundTime(dmean,roundTo=60*60)
        else:
            actime = num2date(dmean)
        return actime

    def processData(self, vals, ntptime):
        """
        processing and storing data - requires sensorid and meta info:
        Writing the following info as data:
        x,y,z,f,dx,dy,t1,t2,var1,var2,var3
        U0,U90,rho,phase,U0error,U90error,4PL voltage,external voltage,current,frequency, N
        MetaInformation:
        SensorID,Softwareversion,
        """
        # currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(ntptime, "%Y-%m-%d")
        filename = outdate
        sensorid = self.sensorid
        header = ''
        datearray = datetime2array(ntptime)
        packcode = '6hLlllllllllll'
        multiplier = [10000,10000,1000,10000,10000,10000,1000,1000,1000,1000,1]

        # convert vals to fullresult
        fullres = [vals[0],vals[1],self.rho(vals[0]/1000.,self.I,self.A,self.L),self.phase(vals[0],vals[1]),self.error(vals[2],vals[0],self.N),self.error(vals[3],vals[0],self.N),vals[4],vals[5],self.I,self.F,self.N]
        #print ("U0[mV]={:.3f}, std={:.3f}, percental error={:.2f}%".format(np.mean(sum0),np.std(sum0),np.std(sum0)/np.mean(sum0)/np.sqrt(29)*100.))
        #print ("U90[mV]={:.3f}, std={:.3f}, percental error={:.2f}%".format(np.mean(sum90),np.std(sum90),np.std(sum90)/np.mean(sum90)/np.sqrt(29)*100. ) )
        #print ("Rho[Ohm*m]={:.1f},Phase[mrad]={:.2f}".format(rho(np.mean(sum0)/1000.,I,A),phase(np.mean(sum0),np.mean(sum90))))

        if self.debug:
            print ("Processing results for {}: {}".format(sensorid, vals))
        if len(fullres) > 3:
            errormark = False
            for idx,el in enumerate(fullres):
                try:
                    datearray.append(int(el*multiplier[idx]))
                except:
                    log.msg('{} protocol: Error while appending data to file - data looks like {}'.format(self.sensordict.get('protocol'),vals))
                    errormark = True
            try:
                data_bin = struct.pack('<'+packcode,*datearray) #little endian
            except:
                log.msg('{} protocol: Error while packing binary data'.format(self.sensordict.get('protocol')))
                errormark = True

            if not errormark:
                #asseble header from available global information - write only if information is complete
                key = '[x,y,z,f,dx,dy,t1,t2,var1,var2,var3]'
                ele = '[U0,U90,Rho,Phase,eU0,eU90,Vint,Vext,I,F,N]'
                unit = '[mV,mV,Ohm*m,mrad,per,per,V,V,uA,Hz,]'
                multplier = "[{}]".format(",".join([str(el) for el in multiplier]))
                # Correct some common old problem
                unit = unit.replace('deg C', 'degC')

                header = "# MagPyBin {} {} {} {} {} {} {}".format(sensorid, key, ele, unit, multplier, packcode, struct.calcsize('<'+packcode))
                data = ','.join(list(map(str,datearray)))

                if not self.confdict.get('bufferdirectory','') == '':
                    mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
            else:
                data = ''
                header = ''
        else:
            data = ''

        return data, header


    def sendRequest(self):

        ser = self.ser
        # connect serial
        if not ser.isOpen():
            ser.open()

        if not self.I > 0:
            log.msg("4PLprotocol: initialization didnt work: current is zero")
            return False
        success = True

        if self.debug:
            print ("Running measurement:")
            print ("-------------------")

        ok = True
        if ok:
            #
            st = datetime.now(timezone.utc).replace(tzinfo=None)
            # start transmission
            comms = ["w","O","S","M","T"]
            sum0,sum90 = [],[]
            if self.debug:
                print ("------------------------")
                print ("Date:", datetime.now(timezone.utc).replace(tzinfo=None))
            for comm in comms:
                if comm == "M":
                    for o in range(0,self.N):
                        answer,actime = send_command(ser,comm,self.eol)
                        answer = answer.replace("M","")
                        #print ("Current {}: U0[mV]={},U90[mV]={}".format(currentdic[curr],answer.split(" ")[0],answer.split(" ")[1]))
                        #print ("U0[mV]={},U90[mV]={}".format(answer.split(" ")[0],answer.split(" ")[1]))
                        u0 = float(answer.split(" ")[0])
                        u90 = float(answer.split(" ")[1])
                        sum0.append(u0)
                        sum90.append(u90)
                        if o == (self.N-1):
                            meanU0=np.mean(sum0)
                            meanU90=np.mean(sum90)
                            stdU0=np.std(sum0)
                            stdU90=np.std(sum90)
                else:
                    answer,actime = send_command(ser,comm,self.eol)
                    if self.debug:
                        print("Receiving {}: {}".format(self.commanddict.get(comm),answer))
                    if comm=="w":
                        Vext=float(answer.replace("w",""))
                    if comm=="O":
                        Vint=float(answer.replace("O",""))
            et = datetime.now(timezone.utc).replace(tzinfo=None)
            actime = self.gettime(st,et)
            result = [meanU0,meanU90,stdU0,stdU90,Vint,Vext]
            if self.debug:
                print ("Processing data: {}".format(result))
            data, head = self.processData(result, actime)
            # send data via mqtt
            if not data == '':
                self.sendmqtt(data,head)

        # disconnect from serial
        #ser.close()


    def sendmqtt(self,data,head):
        """
            call this method after processing data
        """
        sensorid=self.sensorid
        ok = True
        if ok:

            topic = self.confdict.get('station') + '/' + sensorid

            senddata = False
            try:
                coll = int(self.sensordict.get('stack'))
            except:
                coll = 0

            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(data)
                    self.datacnt += 1
                else:
                    senddata = True
                    data = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                if self.debug:
                    print ("publishing...")
                self.client.publish(topic+"/data", data, qos=self.qos)
                # If multiple sensors are connected, self.count needs to be a dict (with topic)
                # Initialize counter for each topic
                if self.counter.get(sensorid,'') == '':
                    self.counter[sensorid] = 0
                cnt = self.counter.get(sensorid)

                if cnt == 0:
                    ## 'Add' is a string containing dict info like:
                    ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,...
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDescription:{},DataTimeProtocol:{}".format( sensorid,self.confdict.get('station','').strip(),self.sensordict.get('pierid','').strip(),self.sensordict.get('protocol','').strip(),self.sensordict.get('sensorgroup','').strip(),self.sensordict.get('sensordesc','').strip(),self.sensordict.get('ptime','').strip() )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)

                cnt += 1
                if cnt >= self.metacnt:
                    cnt = 0
                # update counter in dict
                self.counter[sensorid] = cnt



from __future__ import print_function
from __future__ import absolute_import

# ###################################################################
# Import packages
# ###################################################################

from magpy.stream import DataStream, KEYLIST, NUMKEYLIST, subtractStreams
import struct
from datetime import datetime 
import json
import base64
import binascii

def datetime2array(t):
        return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

## LORA-ZAMG - protocol
##
class lorawanserver(object):
    """
    application/3/node/0018b2200000034a/rx {"applicationID":"3","applicationName":"Temperature-and-Humidity","deviceName":"TITEC-Multisensor","devEUI":"0018b2200000034a","rxInfo":[{"gatewayID":"00800000a0001285","name":"MTCDT_AEPGW2","rssi":-49,"loRaSNR":7.2,"location":{"latitude":48.248422399999995,"longitude":16.3520512,"altitude":0}}],"txInfo":{"frequency":868500000,"dr":5},"adr":true,"fCnt":457,"fPort":1,"data":"QgASEzQVIg/HVA=="}

    content suggestion: appeui, deveui, sensorname, locationname, sensormodell, -> rest beelike

    topic suggestions:
    headline/station/sensor

    ideally, headline is a unique formatidentifier
    e.g.
    loraz/schwarzenbergplatz/adeunis  od 
    warum: so kann man relativ systematisch stationen abfragen
    mobile sensoren ohne festen standort:
    loraz/mobile/adeunis


    """

    def __init__(self):
        """
        """
        print ("  -> Initializing loraWAN server routines ...")
        #self.payload = payload
        #self.topic = topic
        self.topicidentifier = {'startswith':'application','endswith':'rx'}
        self.datakeytranslator = {'tl':['t1','degC'], 'rf':['var1','per'], 'corr':['var5','none']}
        self.identifier = {}
        self.headdict = {}
        self.headstream = {}

    def GetPayload(self, payload, topic):
        loradict = json.loads(payload)
        # convert loradict to headdict (header) and data_bin
        newpayload, sensorid, headline, header =  self.loradict2datastruct(loradict)
        return newpayload, sensorid, headline, header, self.identifier

    def b2v7(self,b1,b2,div):
                val = ((b2 << 8) + b1)/ float(div)
                return val

    def b2v(self,b1,b2,b3,off):
                v = ((b1 << 8) + b2 << 8) + b3
                val = (v/100000. *6.25) - off
                return val

    def loradict2datastruct(self, loradict):
            datakeytranslator = {'tl':['t1','degC'], 'rf':['var1','per'], 'corr':['var5','none'], 'bat':['var4','per']}
            rxdict = loradict.get('rxInfo')[0]
            locdict = rxdict.get('location')
            header = {}

            header['SensorName'] = loradict.get('deviceName','LORA')
            header['SensorDescription'] = loradict.get('applicationName','not specified')
            header['SensorSerialNum'] = loradict.get('devEUI','')
            header['SensorGroup'] = loradict.get('deviceName','LORA')
            sensorid = header['SensorName'][:5].replace('-','') + '_' + header['SensorSerialNum'] + '_0001'
            header['SensorID'] = sensorid
            header['StationID'] = rxdict.get('gatewayID','undefined')
            header['StationName'] = rxdict.get('name','undefined')

            header['StationLongitude'] = locdict.get('longitude','')
            header['StationLatitude'] = locdict.get('latitude','')
            if not locdict.get('longitude','') == '':
                header['StationLocationReference'] = 'WGS84, EPSG: 4326'
            if locdict.get('altitude','') in ['',0,'0']:
                alt = ''
            else:
                alt = locdict.get('altitude')
            header['StationElevation'] = alt
            if not alt == '':
                header['StationElevationRef'] = 'm NN'

            datacode = loradict.get('data')
            # convert to something like datadict = {"tl":21.75,"rf":36.9}
            barray = bytearray(base64.b64decode(datacode))

            print ("Device:", loradict.get('deviceName'))
            print ("Length Bytearray:", len(barray))

            if len(barray) == 10:
                temp = self.b2v(barray[3],barray[4],barray[5],55)
                rf = self.b2v(barray[7],barray[8],barray[9],25)
                datadict = {"tl":temp, "rf":rf}
            elif len(barray) == 7:
                print ("Found Bytearray 7 with code", datacode)
                temp = self.b2v7(barray[1],barray[2],100)
                rf = self.b2v7(barray[3],barray[4],100)
                bat = self.b2v7(barray[5],barray[6],1)
                datadict = {"tl":temp, "rf":rf, "bat":bat}
            else:
                print ("Found Bytearray of length {} with code", len(barray), datacode)
                print ("Payload looks like", loradict)
                temp = 999.0
                rf = -10.0
                datadict = {"tl":temp, "rf":rf}

            keylist, elemlist, unitlist, multilist = [],[],[],[]
            if not loradict.get('DateTime','') == '':
                time = datetime.strptime(loradict.get('DateTime'),"%Y-%m-%dT%H:%M:%S.%fZ")
            elif not loradict.get('DatumSec','') == '':
                time = datetime.strptime(loradict.get('DatumSec'),"%Y-%m-%dT%H:%M:%S.%fZ")
            else:
                time = datetime.now(timezone.utc).replace(tzinfo=None)
            datalst = datetime2array(time)
            packstr = '6hL'
            for elem in datadict:
                if elem in datakeytranslator:
                    key = datakeytranslator[elem][0]
                    unit = datakeytranslator[elem][1]
                    keylist.append(key)
                    elemlist.append(elem)
                    unitlist.append(unit)
                    multilist.append(1000)
                    packstr += "l"
                    datalst.append(int(datadict[elem]*1000))   
                #print (elem, datadict[elem])

            datalst = [str(elem) for elem in datalst]
            dataline =','.join(datalst)
            #print ("DATA", dataline)
            self.identifier[sensorid+':packingcode'] = packstr
            self.identifier[sensorid+':keylist'] = keylist
            self.identifier[sensorid+':elemlist'] = elemlist
            self.identifier[sensorid+':unitlist'] = unitlist
            self.identifier[sensorid+':multilist'] = multilist

            def identifier2line(dic, sensorid):
                p1 = dic.get(sensorid+':packingcode')
                p2 = dic.get(sensorid+':keylist')
                p3 = dic.get(sensorid+':elemlist')
                p4 = dic.get(sensorid+':unitlist')
                p5 = dic.get(sensorid+':multilist')
                p5 = [str(elem) for elem in p5]
                size = struct.calcsize(p1)
                line = "# MagPyBin {} [{}] [{}] [{}] [{}] {} {}".format(sensorid,','.join(p2),','.join(p3),','.join(p4),','.join(p5),p1,size)
                return line

            headline = identifier2line(self.identifier, sensorid)
            #self.headstream[sensorid] = create_head_dict(self.headdict[sensorid],sensorid)
            #self.headstream[sensorid] = merge_two_dicts(self.headstream[sensorid], header)
            #print ("HEAD1", headdict[sensorid])
            #print ("HEAD2", headstream[sensorid])
            print ("success")
            return dataline, sensorid, headline, header


